@@ -1,3 +1,4 @@
+;; -*- lexical-binding: t -*-
 ;;; agda2-mode.el --- Major mode for Agda
 ;; SPDX-License-Identifier: MIT License
 
@@ -28,13 +29,12 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.6.3.1"
+(defvar agda2-version "2.7.20250510"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
 (require 'cl-lib)
 (require 'compile)
-(require 'pp)
 (require 'time-date)
 (require 'eri)
 (require 'annotation)
@@ -99,12 +99,18 @@ argument, and does not need to be listed here."
   :type '(repeat string)
   :group 'agda2)
 
-(defvar agda2-backends '("GHC" "GHCNoMain" "JS" "LaTeX" "QuickLaTeX")
+(defvar agda2-backends '("GHC" "GHCNoMain" "JS" "LaTeX" "QuickLaTeX" "HTML")
   "Compilation backends.")
 
 (defcustom agda2-backend
-  ""
-  "The backend used to compile Agda programs (leave blank to ask every time)."
+  nil
+  "The backend used to compile Agda programs (nil to ask every time)."
+  :type 'string
+  :group 'agda2)
+
+(defcustom agda2-backend-default-payload
+  nil
+  "The default payload for interactive backend commands (nil to ask every time)."
   :type 'string
   :group 'agda2)
 
@@ -219,6 +225,7 @@ constituents.")
     (agda2-load                              "\C-c\C-l"           (global)       "Load")
     (agda2-load                              "\C-c\C-x\C-l")
     (agda2-compile                           "\C-c\C-x\C-c"       (global)       "Compile")
+    (agda2-backend-cmd                       "\C-c\C-i"           (global)       "Run backend interaction")
     (agda2-quit                              "\C-c\C-x\C-q"       (global)       "Quit")
     (agda2-restart                           "\C-c\C-x\C-r"       (global)       "Kill and restart Agda")
     (agda2-abort                             "\C-c\C-x\C-a"       (global)       "Abort a command")
@@ -233,7 +240,7 @@ constituents.")
     (agda2-give                              ,(kbd "C-c C-SPC")   (local)        "Give")
     (agda2-elaborate-give                    ,(kbd "C-c C-m")     (local)        "Elaborate and Give")
     (agda2-refine                            "\C-c\C-r"           (local)        "Refine")
-    (agda2-auto-maybe-all                    "\C-c\C-a"           (local global) "Auto")
+    (agda2-mimer-maybe-all                   "\C-c\C-a"           (local global) "Auto")
     (agda2-make-case                         "\C-c\C-c"           (local)        "Case")
     (agda2-goal-type                         "\C-c\C-t"           (local)        "Goal type")
     (agda2-show-context                      "\C-c\C-e"           (local)        "Context (environment)")
@@ -283,7 +290,7 @@ menus.")
 (defvar agda2-goal-map
   (let ((map (make-sparse-keymap "Agda goal")))
     (dolist (d (reverse agda2-command-table))
-      (cl-destructuring-bind (f &optional keys kinds desc) d
+      (cl-destructuring-bind (f &optional _keys kinds desc) d
         (if (member 'local kinds)
             (define-key map
               (vector (intern desc)) (cons desc f)))))
@@ -530,7 +537,7 @@ process."
 Sends the list of strings ARGS to the Agda2 interpreter, waits
 for output and executes the responses, if any.
 
-If SAVE is 'save, then the buffer is saved first.
+If SAVE is \\='save, then the buffer is saved first.
 
 If HIGHLIGHT is non-nil, then the buffer's syntax highlighting
 may be updated. This is also the case if the Agda process is
@@ -609,7 +616,7 @@ successful."
   (setq agda2-highlight-in-progress nil
         agda2-last-responses        nil))
 
-(defun agda2-output-filter (proc chunk)
+(defun agda2-output-filter (_proc chunk)
   "Evaluate the Agda process's commands.
 This filter function assumes that every line contains either some
 kind of error message (which cannot be parsed as a list), or
@@ -759,13 +766,13 @@ The user input is computed as follows:
   contains whitespace, then the input is taken from the
   minibuffer. In this case WANT is used as the prompt string.
 
-* Otherwise (including if WANT is 'goal) the goal contents are
+* Otherwise (including if WANT is \\='goal) the goal contents are
   used.
 
 If the user input is not taken from the goal, then an empty goal
 range is given.
 
-If SAVE is 'save, then the buffer is saved just before the
+If SAVE is \\='save, then the buffer is saved just before the
 command is sent to Agda (if it is sent)."
   (cl-multiple-value-bind (o g) (agda2-goal-at (point))
     (unless g (error "For this command, please place the cursor in a goal"))
@@ -804,24 +811,37 @@ command is sent to Agda (if it is sent)."
             (agda2-list-quote agda2-program-args)
             ))
 
-(defun agda2-compile ()
-  "Compile the current module.
+(defun agda2-read-backend ()
+  "Get the currently set backend from the `agda2-backend' variable,
+or ask the user for input in case the variable is not set."
+  (let ((name (cond
+    ((equal agda2-backend "MAlonzo")       "GHC")
+    ((equal agda2-backend "MAlonzoNoMain") "GHCNoMain")
+    ((equal agda2-backend nil)
+      (completing-read "Backend: " agda2-backends nil nil nil nil nil 'inherit-input-method))
+    (t agda2-backend))))
+  (when (equal name "") (error "No backend chosen"))
+  name))
+
+(defun agda2-compile (backend)
+  "Compile the current module using BACKEND.
 
 The variable `agda2-backend' determines which backend is used."
-  (interactive)
-  (let ((backend (cond ((equal agda2-backend "MAlonzo")       "GHC")
-                       ((equal agda2-backend "MAlonzoNoMain") "GHCNoMain")
-                       ((equal agda2-backend "")
-                        (completing-read "Backend: " agda2-backends
-                                         nil nil nil nil nil
-                                         'inherit-input-method))
-                       (t agda2-backend))))
-    (when (equal backend "") (error "No backend chosen"))
-    (agda2-go 'save t 'busy t "Cmd_compile"
-              backend
-              (agda2-string-quote (buffer-file-name))
-              (agda2-list-quote agda2-program-args)
-              )))
+  (interactive (list (agda2-read-backend)))
+  (agda2-go 'save t 'busy t "Cmd_compile"
+            backend
+            (agda2-string-quote (buffer-file-name))
+            (agda2-list-quote agda2-program-args)
+            ))
+
+(defun agda2-backend-cmd (backend payload)
+  "Run the current backend's (from variable `agda2-backend`) interaction command."
+  (interactive (list (agda2-read-backend) (or agda2-backend-default-payload (read-string "Payload: "))))
+  (if (agda2-goal-at (point))
+     (agda2-goal-cmd "Cmd_backend_hole" nil 'goal nil backend (agda2-string-quote payload))
+     (agda2-go 'save t 'busy t "Cmd_backend_top" backend (agda2-string-quote payload))
+  )
+)
 
 (defmacro agda2-maybe-forced (name comment cmd save want)
   "This macro constructs a function NAME which runs CMD.
@@ -887,16 +907,6 @@ of new goals."
       (agda2-goal-cmd "Cmd_refine_or_intro True" 'save 'goal)
     (agda2-goal-cmd "Cmd_refine_or_intro False" 'save 'goal)))
 
-(defun agda2-autoOne ()
- "Simple proof search" (interactive)
- (agda2-goal-cmd "Cmd_autoOne" 'save 'goal))
-
-(defun agda2-autoAll ()
-  (interactive)
-  "Solves all goals by simple proof search."
-  (agda2-go nil nil 'busy t "Cmd_autoAll")
-)
-
 (defun agda2-make-case ()
   "Refine the pattern variables given in the goal.
 Assumes that <clause> = {!<variables>!} is on one line."
@@ -963,8 +973,8 @@ The buffer is returned.")
       (compilation-mode "AgdaInfo")
       ;; Support for jumping to positions mentioned in the text.
       (set (make-local-variable 'compilation-error-regexp-alist)
-           '(("\\([\\\\/][^[:space:]]*\\):\\([0-9]+\\),\\([0-9]+\\)-\\(\\([0-9]+\\),\\)?\\([0-9]+\\)"
-              1 (2 . 5) (3 . 6))))
+           '(("\\([\\\\/][^[:space:]]*\\):\\([0-9]+\\)\\.\\([0-9]+\\)\\(-\\(\\([0-9]+\\)\\.\\)?\\([0-9]+\\)\\)?"
+              1 (2 . 6) (3 . 7))))
       ;; Do not skip errors that start in the same position as the
       ;; current one.
       (set (make-local-variable 'compilation-skip-to-next-location) nil)
@@ -1131,7 +1141,7 @@ commands have completed."
     (when (buffer-live-p agda2-process-buffer)
       (kill-buffer agda2-process-buffer))))
 
-(defun agda2-kill-process-buffer (&optional process event)
+(defun agda2-kill-process-buffer (&optional _process _event)
   "Kills the Agda process buffer, if any.
 But only if the Agda process does not exist or has terminated.
 
@@ -1188,7 +1198,7 @@ The form of the result depends on the prefix argument:
        ;; Those called from a goal, grab the value present there (if any)
        ;; Whereas those called globally always use a prompt
        (interactive ,(pcase spec
-                       (`(fromgoal ,want)
+                       (`(fromgoal ,_want)
                         "P")
                        (`(global ,prompt)
                         (if prompt
@@ -1380,13 +1390,26 @@ Either only one if point is a goal, or all of them."
                           'agda2-solveAll))
 )
 
-(defun agda2-auto-maybe-all ()
-  "Run auto.
+(defun agda2-mimer-maybe-all ()
+  "Run proof search.
 Either only one if point is a goal, or all of them."
   (interactive)
   (call-interactively (if (agda2-goal-at (point))
-                          'agda2-autoOne
-                          'agda2-autoAll))
+                          'agda2-mimer
+                          'agda2-mimerAll))
+)
+
+(agda2-maybe-normalised-asis
+  agda2-mimer
+  "Run proof search on a goal."
+  "Cmd_autoOne"
+  'goal
+)
+
+(agda2-maybe-normalised-toplevel-asis-noprompt
+  agda2-mimerAll
+  "Solves all goals by simple proof search."
+  "Cmd_autoAll"
 )
 
 (agda2-maybe-normalised-toplevel-asis-noprompt
@@ -1521,7 +1544,7 @@ ways."
        ;; Don't run modification hooks: we don't want this function to
        ;; trigger agda2-abort-highlighting.
        (inhibit-modification-hooks t))
-      ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--\\|^%.*\\\\begin{code}\\|\\\\begin{code}\\|\\\\end{code}\\|```\\|\\#\\+begin_src agda2\\|\\#\\+end_src agda2" nil t))
+      ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--\\|^%.*\\\\begin{code}\\|\\\\begin{code}\\|\\\\end{code}\\|```\\|\\#\\+begin_src agda2\\|\\#\\+end_src" nil t))
        ;; is-proper checks whether string s (e.g. "?" or "--") is proper
        ;; i.e., is not part of an identifier.
        ;; comment-starter is true if s starts a comment (e.g. "--")
@@ -1557,7 +1580,7 @@ ways."
           ("\\begin{code}"     (when (outside-code)               (pop stk)))
           ("\\end{code}"       (when (not stk)                    (push 'outside stk)))
           ("#+begin_src agda2" (when (outside-code)               (pop stk)))
-          ("#+end_src agda2"   (when (not stk)                    (push 'outside stk)))
+          ("#+end_src"         (when (not stk)                    (push 'outside stk)))
           ("```"               (if   (outside-code)               (pop stk)
                                (when (not stk)                    (push 'outside stk))))
           ("--"                (when (and (not stk)
@@ -1590,7 +1613,7 @@ ways."
      (overlay-put o 'face               'highlight)
      (overlay-put o 'after-string       (propertize (format "%s" n) 'face 'highlight)))))
 
-(defun agda2-protect-goal-markers (ol action beg end &optional length)
+(defun agda2-protect-goal-markers (ol action beg end &optional _length)
   "Ensures that the goal markers cannot be tampered with.
 Except if `inhibit-read-only' is non-nil or /all/ of the goal is
 modified."
@@ -1649,7 +1672,12 @@ text properties."
     (error "no process")))
 
 (defun agda2-intersperse (sep xs)
-  (let(ys)(while xs (push (pop xs) ys)(push sep ys))(pop ys)(nreverse ys)))
+  (let (ys)
+    (while xs
+      (push (pop xs) ys)
+      (push sep ys))
+    (pop ys)
+    (nreverse ys)))
 
 (defun agda2-goal-Range (o)
   "The Haskell Range of goal overlay O."
@@ -1662,7 +1690,7 @@ POINTS must be a list of integers, and its length must be 0 or 2."
   (if points
       (format "(intervalsToRange (Just (mkAbsolute %s)) %s)"
               (agda2-string-quote (file-truename (buffer-file-name)))
-              (format "[Interval %s %s]"
+              (format "[Interval () %s %s]"
                       (agda2-mkPos (car points))
                       (agda2-mkPos (cadr points))))
     "noRange"))
@@ -1694,10 +1722,10 @@ points < 128) are converted to singleton strings."
 Removes any text properties, escapes newlines, double quotes,
 etc., adds surrounding double quotes, and converts non-ASCII
 characters to the \\xNNNN notation used in Haskell strings."
-  (let ((pp-escape-newlines t)
+  (let ((print-escape-newlines t)
         (s2 (copy-sequence s)))
     (set-text-properties 0 (length s2) nil s2)
-    (mapconcat 'agda2-char-quote (pp-to-string s2) "")))
+    (mapconcat 'agda2-char-quote (prin1-to-string s2) "")))
 
 (defun agda2-list-quote (strings)
   "Convert a list of STRINGS into a string representing it in Haskell syntax."
@@ -1948,23 +1976,37 @@ the argument is a positive number, otherwise turn it off."
 ;; Switching to a different version of Agda
 
 (defun agda2-get-agda-program-versions ()
-  "Get \"version strings\" of executables starting with
-'agda-mode' in current path."
+  "Get suffixes (\"version strings\") of executables starting with
+\\='agda-\\=' in exec-path."
+  (interactive) ;; for debugging purposes
   (delete-dups
    (mapcar (lambda (path)
-             ;; strip 'agda-mode' prefix
-             (replace-regexp-in-string "^agda-mode-?" ""
+             ;; strip 'agda-' prefix
+             (replace-regexp-in-string "^agda-?" ""
                                        (file-name-nondirectory path)))
            (cl-remove-if-not 'file-executable-p
              ;; concatenate result
              (cl-reduce 'append
                      ;; for each directory in exec-path, get list of
-                     ;; files whose name starts with 'agda-mode'
+                     ;; files whose name starts with 'agda-'
                      (mapcar (lambda (path)
                                (when (file-accessible-directory-p path)
-                                 (directory-files path 't "^agda-mode")))
+                                 (directory-files path 't "^agda-")))
                              exec-path))))))
 
+
+(defun agda2-old-version (version)
+  "Check if the given version is less that 2.8.
+
+Agda 2.8 is the version where the agda-mode executable was deprecated.
+
+If the version cannot be parsed, the answer is no.
+So e.g. agda-fast would be interpreted as not older than
+agda-2.7.20250510"
+  ;; If the given string does not have a numeric prefix, we answer no.
+  (and (string-match "^[0-9][0-9.]*" version)
+       (version< (match-string 0 version) "2.8")))
+
 ;; Note that other versions of Agda may use different protocols, so
 ;; this function unloads the Emacs mode.
 
@@ -1972,12 +2014,21 @@ the argument is a positive number, otherwise turn it off."
   "Tries to switch to Agda version VERSION.
 
 This command assumes that the agda and agda-mode executables for
-Agda version VERSION are called agda-VERSION and
-agda-mode-VERSION, and that they are located on the PATH. (If
-VERSION is empty, then agda and agda-mode are used instead.)
+Agda version VERSION are called agda-VERSION and agda-mode-VERSION.
+It searches for these executables on the exec-path
+which is often identical to PATH but can be different
+depending how Emacs is configured and invoked.
 
-An attempt is made to preserve the default value of
-`agda2-mode-hook'."
+Only if VERSION is numeric and less than \"2.8\"
+the agda-mode executable is used to locate the respective agda2.el,
+otherwise \"agda --emacs-mode locate\" is invoked.
+
+If VERSION is empty or \"agda\", then agda is used.
+A prefix \"agda-\" is stripped from VERSION,
+so one can invoke this function e.g. with
+\"agda-2.7.20250510\" instead of \"2.7.20250510\".
+
+An attempt is made to preserve the default value of `agda2-mode-hook'."
   (interactive
    (list (completing-read "Version: " (agda2-get-agda-program-versions))))
 
@@ -1991,6 +2042,22 @@ An attempt is made to preserve the default value of
 
        (default-hook (default-value 'agda2-mode-hook))
 
+       ;; Sanitize version.
+       ;; * Strip "agda-" prefix from version.
+       ;; * Replace version "agda" by just "".
+       ;;
+       ;; Rationale:
+       ;; The user might by accident enter "agda-2.7.20250510" instead of "2.7.20250510",
+       ;; or "agda" instead of "".
+       (version
+        (cond
+         ;; Strip "agda-" prefix from version.
+         ((string-prefix-p "agda-" version) (substring version (length "agda-")))
+         ;; Replace version "agda" by just "".
+         ((equal version "agda") "")
+         ;; Default.
+         (t version)))
+
        (version-suffix (if (or (equal version "")
                                (equal version nil))
                            ""
@@ -1999,14 +2066,16 @@ An attempt is made to preserve the default value of
        ;; Run agda-mode<version-suffix> and make sure that it returns
        ;; successfully.
        (coding-system-for-read 'utf-8)
-       (agda-mode-prog (concat "agda-mode" version-suffix))
+       (old-agda (agda2-old-version version))
+       (agda-mode-prog (concat (if old-agda "agda-mode" "agda") version-suffix))
        (agda-mode-path
         (condition-case nil
             (with-temp-buffer
               (unless
-                  (equal 0 (call-process agda-mode-prog
-                                         nil (current-buffer) nil
-                                         "locate"))
+                  (equal 0
+                    (if old-agda
+                        (call-process agda-mode-prog nil (current-buffer) nil "locate")
+                        (call-process agda-mode-prog nil (current-buffer) nil "--emacs-mode" "locate")))
                 (error "%s" (concat "Error when running "
                                     agda-mode-prog)))
               (buffer-string))
