@@ -7,7 +7,6 @@ module Agda.Compiler.MAlonzo.Compiler
 
 import Control.Arrow (second)
 import Control.DeepSeq
-import Control.Monad
 import Control.Monad.Except   ( throwError )
 import Control.Monad.IO.Class ( MonadIO(..) )
 import Control.Monad.Reader   ( MonadReader(..), asks, ReaderT, runReaderT, withReaderT)
@@ -61,7 +60,7 @@ import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Primitive (getBuiltinName)
-import Agda.TypeChecking.Pretty hiding ((<>))
+import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Warnings
@@ -79,7 +78,7 @@ import Agda.Utils.Monad
 import Agda.Utils.Singleton
 import qualified Agda.Utils.IO.UTF8 as UTF8
 
-import Paths_Agda
+import Agda.Setup ( getDataDir )
 
 import Agda.Utils.Impossible
 
@@ -102,6 +101,8 @@ ghcBackend' = Backend'
   , compileDef            = ghcCompileDef
   , scopeCheckingSuffices = False
   , mayEraseType          = ghcMayEraseType
+  , backendInteractTop    = Nothing
+  , backendInteractHole   = Nothing
   }
 
 --- Command-line flags ---
@@ -218,14 +219,9 @@ data GHCDefinition = GHCDefinition
 
 ghcPreCompile :: GHCFlags -> TCM GHCEnv
 ghcPreCompile flags = do
-  cubical <- optCubical <$> pragmaOptions
-  let notSupported s =
-        typeError $ GenericError $
-          "Compilation of code that uses " ++ s ++ " is not supported."
-  case cubical of
-    Nothing      -> return ()
-    Just CErased -> return ()
-    Just CFull   -> notSupported "--cubical"
+  whenJustM cubicalOption \case
+    CErased -> pure ()
+    CFull   -> typeError $ CubicalCompilationNotSupported CFull
 
   outDir <- compileDir
   let ghcOpts = GHCOptions
@@ -263,8 +259,6 @@ ghcPreCompile flags = do
   mpathp      <- getBuiltinName builtinPathP
   msub        <- getBuiltinName builtinSub
   msubin      <- getBuiltinName builtinSubIn
-  mid         <- getBuiltinName builtinId
-  mconid      <- getPrimitiveName' builtinConId
 
   istcbuiltin <- do
     builtins <- mapM getBuiltinName
@@ -305,9 +299,12 @@ ghcPreCompile flags = do
       , builtinAgdaTCMFormatErrorParts
       , builtinAgdaTCMDebugPrint
       , builtinAgdaTCMNoConstraints
+      , builtinAgdaTCMWorkOnTypes
       , builtinAgdaTCMRunSpeculative
       , builtinAgdaTCMExec
+      , builtinAgdaTCMCheckFromString
       , builtinAgdaTCMGetInstances
+      , builtinAgdaTCMSolveInstances
       , builtinAgdaTCMPragmaForeign
       , builtinAgdaTCMPragmaCompile
       , builtinAgdaBlocker
@@ -352,8 +349,6 @@ ghcPreCompile flags = do
     , ghcEnvPathP       = mpathp
     , ghcEnvSub         = msub
     , ghcEnvSubIn       = msubin
-    , ghcEnvId          = mid
-    , ghcEnvConId       = mconid
     , ghcEnvIsTCBuiltin = istcbuiltin
     , ghcEnvListArity   = listArity
     , ghcEnvMaybeArity  = maybeArity
@@ -365,8 +360,9 @@ ghcPostCompile _cenv _isMain mods = do
   -- FIXME: @curMName@ and @curIF@ are evil TCM state, but there does not appear to be
   --------- another way to retrieve the compilation root ("main" module or interaction focused).
   rootModuleName <- curMName
-  rootModule <- ifJust (Map.lookup rootModuleName mods) pure
-                $ genericError $ "Module " <> prettyShow rootModuleName <> " was not compiled!"
+  -- Mario, 2024-10-16: cannot trigger this error:
+  -- genericError $ "Module " <> prettyShow rootModuleName <> " was not compiled!"
+  let rootModule = Map.findWithDefault __IMPOSSIBLE__ rootModuleName mods
   flip runReaderT rootModule $ do
     copyRTEModules
     callGHC
@@ -382,7 +378,7 @@ ghcPreModule
                  -- ^ Could we confirm the existence of a main function?
 ghcPreModule cenv isMain m mifile =
   (do let check = ifM uptodate noComp yesComp
-      cubical <- optCubical <$> pragmaOptions
+      cubical <- cubicalOption
       case cubical of
         -- Code that uses --cubical is not compiled.
         Just CFull   -> noComp
@@ -407,7 +403,7 @@ ghcPreModule cenv isMain m mifile =
     yesComp = do
       m   <- prettyShow <$> curMName
       out <- curOutFile
-      reportSLn "compile.ghc" 1 $ repl [m, ifileDesc, out] "Compiling <<0>> in <<1>> to <<2>>"
+      alwaysReportSLn "compile.ghc" 1 $ repl [m, ifileDesc, out] "Compiling <<0>> in <<1>> to <<2>>"
       asks Recompile
 
 ghcPostModule
@@ -437,9 +433,9 @@ ghcPostModule _cenv menv _isMain _moduleName ghcDefs = do
     hsModuleName <- curHsMod
     writeModule $ HS.Module
       hsModuleName
-      (map HS.OtherPragma headerPragmas)
+      (map HS.OtherPragma $ List.nub headerPragmas)
       imps
-      (map fakeDecl (hsImps ++ code) ++ decls)
+      (map fakeDecl (List.nub hsImps ++ code) ++ decls)
 
   return $ GHCModule menv mainDefs
 
@@ -463,7 +459,7 @@ ghcMayEraseType q = getHaskellPragma q <&> \case
 -- Compilation ------------------------------------------------------------
 
 imports ::
-  BuiltinThings PrimFun -> Set TopLevelModuleName -> [Definition] ->
+  BuiltinThings -> Set TopLevelModuleName -> [Definition] ->
   [HS.ImportDecl]
 imports builtinThings usedModules defs = hsImps ++ imps where
   hsImps :: [HS.ImportDecl]
@@ -516,9 +512,6 @@ mazRTEFloatImport (UsesFloat b) = [ HS.ImportDecl mazRTEFloat True Nothing | b ]
 
 definition :: Definition -> HsCompileM (UsesFloat, [HS.Decl], Maybe CheckedMainFunctionDef)
 -- ignore irrelevant definitions
-{- Andreas, 2012-10-02: Invariant no longer holds
-definition kit (Defn NonStrict _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
--}
 definition Defn{defArgInfo = info, defName = q} | not $ usableModality info = do
   reportSDoc "compile.ghc.definition" 10 $
     ("Not compiling" <+> prettyTCM q) <> "."
@@ -537,11 +530,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
   (uncurry (,,typeCheckedMainDef)) . second ((mainDecl ++) . infodecl q) <$>
     case d of
 
-      _ | Just (HsDefn r hs) <- pragma -> setCurrentRange r $
-          if is ghcEnvFlat
-          then genericError
-                "\"COMPILE GHC\" pragmas are not allowed for the FLAT builtin."
-          else do
+      _ | Just (HsDefn r hs) <- pragma -> setCurrentRange r $ do
             -- Make sure we have imports for all names mentioned in the type.
             hsty <- haskellType q
             mapM_ (`xqual` HS.Ident "_") (namesIn ty :: Set QName)
@@ -567,8 +556,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
       -- Compiling List
       Datatype{ dataPars = np } | is ghcEnvList -> do
         sequence_ [primNil, primCons] -- Just to get the proper error for missing NIL/CONS
-        caseMaybe pragma (return ()) $ \ p -> setCurrentRange p $ warning . GenericWarning =<< do
-          fsep $ pwords "Ignoring GHC pragma for builtin lists; they always compile to Haskell lists."
+        whenJust pragma $ \ p -> setCurrentRange p $ warning PragmaCompileList
         let d = dname q
             t = unqhname TypeK q
         Just nil  <- getBuiltinName builtinNil
@@ -582,8 +570,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
       -- Compiling Maybe
       Datatype{ dataPars = np } | is ghcEnvMaybe -> do
         sequence_ [primNothing, primJust] -- Just to get the proper error for missing NOTHING/JUST
-        caseMaybe pragma (return ()) $ \ p -> setCurrentRange p $ warning . GenericWarning =<< do
-          fsep $ pwords "Ignoring GHC pragma for builtin maybe; they always compile to Haskell lists."
+        whenJust pragma $ \ p -> setCurrentRange p $ warning PragmaCompileMaybe
         let d = dname q
             t = unqhname TypeK q
         Just nothing <- getBuiltinName builtinNothing
@@ -698,39 +685,6 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
                  emptyBinds]
           ]
 
-      -- Id x y is compiled as a pair of a boolean and whatever
-      -- Path x y is compiled to.
-      Datatype{} | is ghcEnvId -> do
-        sequence_ [primInterval]
-        Just int <- getBuiltinName builtinInterval
-        int      <- xhqn TypeK int
-        -- re  #3733: implement reflId
-        retDecls $
-          [ HS.TypeDecl (unqhname TypeK q)
-              [] -- [HS.UnkindedVar (ihname A i) | i <- [0..3]]
-              (HS.TyApp (HS.FakeType "(,) Bool")
-                 (HS.TyFun (HS.TyCon int) mazAnyType))
-          , HS.FunBind
-              [HS.Match (dname q) []
-                 (HS.UnGuardedRhs (HS.FakeExp "\\_ _ _ _ -> ()"))
-                 emptyBinds]
-          ]
-
-      -- conid.
-      Primitive{} | is ghcEnvConId -> do
-        strict <- optGhcStrictData <$> askGhcOpts
-        let var = applyWhen strict HS.PBangPat . HS.PVar
-        retDecls $
-          [ HS.FunBind
-              [HS.Match (dname q)
-                 [ var (ihname A i) | i <- [0..1] ]
-                 (HS.UnGuardedRhs $
-                  HS.App (HS.App (HS.FakeExp "(,)")
-                            (HS.Var (HS.UnQual (ihname A 0))))
-                    (HS.Var (HS.UnQual (ihname A 1))))
-                 emptyBinds]
-          ]
-
       -- TC builtins are compiled to erased, which is an ∞-ary
       -- function.
       Axiom{} | ghcEnvIsTCBuiltin env q -> do
@@ -803,12 +757,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
   function mhe fun = do
     (imp, ccls) <- fun
     case mhe of
-      Just (HsExport r name) -> setCurrentRange r $ do
-        env <- askGHCEnv
-        if Just q == ghcEnvFlat env
-        then genericError
-              "\"COMPILE GHC as\" pragmas are not allowed for the FLAT builtin."
-        else do
+      Just (HsExport r name) -> do
           t <- setCurrentRange r $ haskellType q
           let tsig :: HS.Decl
               tsig = HS.TypeSig [HS.Ident name] t
@@ -877,7 +826,9 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
 
 constructorCoverageCode :: QName -> Int -> [QName] -> HaskellType -> [HaskellCode] -> HsCompileM [HS.Decl]
 constructorCoverageCode q np cs hsTy hsCons = do
-  liftTCM $ checkConstructorCount q cs hsCons
+  -- Check that number of constructors matches up.
+  unless (length cs == length hsCons) $
+    ghcBackendError $ ConstructorCountMismatch q cs hsCons
   ifM (liftTCM $ noCheckCover q) (return []) $ do
     ccs <- List.concat <$> zipWithM checkConstructorType cs hsCons
     cov <- liftTCM $ checkCover q hsTy np cs hsCons
@@ -1341,9 +1292,7 @@ callGHC = do
   let isMain = modIsMain && modHasMainFunc  -- both need to be IsMain
 
   -- Warn if no main function and not --no-main
-  when (modIsMain /= isMain) $
-    genericWarning =<< fsep (pwords "No main function defined in" ++ [prettyTCM agdaMod <> "."] ++
-                             pwords "Use --no-main to suppress this warning.")
+  when (modIsMain /= isMain) $ warning $ NoMain agdaMod
 
   let overridableArgs =
         [ "-O"] ++
