@@ -1,5 +1,4 @@
 {-# LANGUAGE NondecreasingIndentation #-}
-
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.Interaction.BasicOps where
@@ -7,10 +6,9 @@ module Agda.Interaction.BasicOps where
 import Prelude hiding (null)
 
 import Control.Arrow          ( first )
-import Control.Monad          ( (<=<), (>=>), forM, filterM, guard )
-import Control.Monad.Except
-import Control.Monad.State
-import Control.Monad.Identity
+import Control.Monad.Except   ( MonadError(..) )
+import Control.Monad.State    ( MonadState(..), evalState )
+import Control.Monad.Identity ( runIdentity )
 import Control.Monad.Trans.Maybe
 
 import qualified Data.Map as Map
@@ -25,6 +23,7 @@ import Data.Text (Text)
 import qualified Data.Text as T
 
 import Agda.Interaction.Base
+import Agda.Interaction.Output
 import Agda.Interaction.Options
 import Agda.Interaction.Response (Goals, ResponseContextEntry(..))
 
@@ -49,7 +48,7 @@ import Agda.Syntax.Parser
 import Agda.TheTypeChecker
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Errors ( getAllWarnings, stringTCErr, Verbalize(..) )
+import Agda.TypeChecking.Errors ( getAllWarnings, Verbalize(..) )
 import Agda.TypeChecking.Monad as M hiding (MetaInfo)
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.MetaVars.Mention
@@ -71,8 +70,8 @@ import Agda.TypeChecking.CheckInternal
 import Agda.TypeChecking.SizedTypes.Solve
 import qualified Agda.TypeChecking.Pretty as TP
 import Agda.TypeChecking.Warnings
-  ( runPM, warning, WhichWarnings(..), classifyWarnings, isMetaTCWarning
-  , WarningsAndNonFatalErrors, emptyWarningsAndNonFatalErrors )
+  ( warning, WhichWarnings(..), classifyWarnings, isMetaTCWarning
+  , WarningsAndNonFatalErrors )
 
 import Agda.Termination.TermCheck (termMutual)
 
@@ -100,8 +99,7 @@ parseExpr rng s = do
   (C.ExprWhere e wh, attrs) <-
     runPM $ parsePosString exprWhereParser pos s
   checkAttributes attrs
-  unless (null wh) $ typeError $ GenericError $
-    "where clauses are not supported in holes"
+  unless (null wh) $ interactionError UnexpectedWhere
   return e
   where pos = fromMaybe (startPos Nothing) $ rStart rng
 
@@ -149,7 +147,7 @@ giveExpr force mii mi e = do
       reportSDoc "interaction.give" 40 $ "give: checked expression:" TP.<+> pure (pretty v)
       case mvInstantiation mv of
 
-        InstV{} -> unlessM ((Irrelevant ==) <$> viewTC eRelevance) $ do
+        InstV{} -> unlessM (isIrrelevant <$> viewTC eRelevance) $ do
           v' <- instantiate $ MetaV mi $ map Apply ctx
           reportSDoc "interaction.give" 20 $ TP.sep
             [ "meta was already set to value v' = " TP.<+> prettyTCM v'
@@ -167,7 +165,7 @@ giveExpr force mii mi e = do
       unless (force == WithForce) $ redoChecks mii
       wakeupConstraints mi
       solveSizeConstraints DontDefaultToInfty
-      cubical <- isJust . optCubical <$> pragmaOptions
+      cubical <- isJust <$> cubicalOption
       -- don't double check with cubical, because it gets in the way too often.
       unless (cubical || force == WithForce) $ do
         -- Double check.
@@ -186,36 +184,46 @@ redoChecks (Just ii) = do
   case ipClause ip of
     IPNoClause -> return ()
     IPClause{ipcQName = f} -> do
-      mb <- mutualBlockOf f
+      mb <- defMutual <$> getConstInfo f
       terErrs <- localTC (\ e -> e { envMutualBlock = Just mb }) $ termMutual []
-      unless (null terErrs) $ warning $ TerminationIssue terErrs
+      List1.unlessNull terErrs $ warning . TerminationIssue
   -- TODO redo positivity check!
 
--- | Try to fill hole by expression.
---
---   Returns the given expression unchanged
---   (for convenient generalization to @'refine'@).
-give
-  :: UseForce       -- ^ Skip safety checks?
+-- | Auxiliary definition for 'give' and 'elaborate_give'.
+give_ ::
+     Bool           -- ^ Elaborating?
+  -> UseForce       -- ^ Skip safety checks?
   -> InteractionId  -- ^ Hole.
   -> Maybe Range
   -> Expr           -- ^ The expression to give.
-  -> TCM Expr       -- ^ If successful, the very expression is returned unchanged.
-give force ii mr e = liftTCM $ do
+  -> TCM Term       -- ^ Value of the expression
+give_ elaborating force ii mr e = do
   -- if Range is given, update the range of the interaction meta
   mi  <- lookupInteractionId ii
   whenJust mr $ updateMetaVarRange mi
   reportSDoc "interaction.give" 10 $ "giving expression" TP.<+> prettyTCM e
   reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
   -- Try to give mi := e
-  _ <- withInteractionId ii $ do
-     setMetaOccursCheck mi DontRunMetaOccursCheck -- #589, #2710: Allow giving recursive solutions.
-     giveExpr force (Just ii) mi e
-    `catchError` \ case
-      -- Turn PatternErr into proper error:
-      PatternErr{} -> typeError . GenericDocError =<< do
-        withInteractionId ii $ "Failed to give" TP.<+> prettyTCM e
-      err -> throwError err
+  withInteractionId ii do
+    setMetaOccursCheck mi DontRunMetaOccursCheck -- #589, #2710: Allow giving recursive solutions.
+    applyWhen elaborating (locallyTC eCurrentlyElaborating $ const True) do
+      -- Andreas, 2025-05-02, issue #7842 reproduces this error
+      let err _blocker = typeError $ InteractionError $ CannotGive e
+      catchPatternErr err do
+        giveExpr force (Just ii) mi e
+
+-- | Try to fill hole by expression.
+--
+--   Returns the given expression unchanged
+--   (for convenient generalization to @'refine'@).
+give ::
+     UseForce       -- ^ Skip safety checks?
+  -> InteractionId  -- ^ Hole.
+  -> Maybe Range
+  -> Expr           -- ^ The expression to give.
+  -> TCM Expr       -- ^ If successful, the very expression is returned unchanged.
+give force ii mr e = do
+  _ <- give_ False force ii mr e
   removeInteractionPoint ii
   return e
 
@@ -228,26 +236,11 @@ elaborate_give
   -> Expr           -- ^ The expression to give.
   -> TCM Expr       -- ^ If successful, return the elaborated expression.
 elaborate_give norm force ii mr e = withInteractionId ii $ do
-  -- if Range is given, update the range of the interaction meta
-  mi  <- lookupInteractionId ii
-  whenJust mr $ updateMetaVarRange mi
-  reportSDoc "interaction.give" 10 $ "giving expression" TP.<+> prettyTCM e
-  reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
-  -- Try to give mi := e
-  v <- withInteractionId ii $ do
-     setMetaOccursCheck mi DontRunMetaOccursCheck -- #589, #2710: Allow giving recursive solutions.
-     locallyTC eCurrentlyElaborating (const True) $
-       giveExpr force (Just ii) mi e
-    `catchError` \ case
-      -- Turn PatternErr into proper error:
-      PatternErr{} -> typeError . GenericDocError =<< do
-        withInteractionId ii $ "Failed to give" TP.<+> prettyTCM e
-      err -> throwError err
-  mv <- lookupLocalMeta mi
+  v <- give_ True force ii mr e
+  reportSDoc "interaction.give" 40 $ "v = " TP.<+> pure (pretty v)
   -- Reduce projection-likes before quoting, otherwise instance
   -- selection may fail on reload (see #6203).
   nv <- reduceProjectionLike =<< normalForm norm v
-  reportSDoc "interaction.give" 40 $ "nv = " TP.<+> pure (pretty v)
   locallyTC ePrintMetasBare (const True) $ reify nv
 
 -- | Try to refine hole by expression @e@.
@@ -276,11 +269,10 @@ refine force ii mr e = do
     tryRefine nrOfMetas r scope = try nrOfMetas Nothing
       where
         try :: Int -> Maybe TCErr -> Expr -> TCM Expr
-        try 0 err e = throwError . stringTCErr $ case err of
+        try 0 err e = interactionError $ CannotRefine $ case err of
            Just (TypeError _ _ cl) | UnequalTerms _ I.Pi{} _ _ <- clValue cl ->
-             "Cannot refine functions with 10 or more arguments"
-           _ ->
-             "Cannot refine"
+             "functions with 10 or more arguments"
+           _ -> ""
         try n _ e = give force ii (Just r) e `catchError` \err -> try (n - 1) (Just err) =<< appMeta e
 
         -- Apply A.Expr to a new meta
@@ -297,6 +289,7 @@ refine force ii mr e = do
                     -- should be something else.
                 , metaNumber = Nothing -- in order to print just as ?, not ?n
                 , metaNameSuggestion = ""
+                , metaKind           = Info.UnificationMeta
                 }
               metaVar = QuestionMark info ii
 
@@ -319,20 +312,6 @@ refine force ii mr e = do
                   _ -> App i e arg
           return $ smartApp (defaultAppInfo r) e $ defaultNamedArg metaVar
 
--- Andreas, 2017-12-16:
--- Ulf, your attempt to fix #737 introduced regression #2873.
--- Going through concrete syntax does some arbitrary disambiguation
--- of constructors, which subsequently makes refine fail.
--- I am not convinced of the printing-parsing shortcut to address problems.
--- (Unless you prove the roundtrip property.)
---
---           rescopeExpr scope $ smartApp (defaultAppInfo r) e $ defaultNamedArg metaVar
--- -- | Turn an abstract expression into concrete syntax and then back into
--- --   abstract. This ensures that context precedences are set correctly for
--- --   abstract expressions built by hand. Used by refine above.
--- rescopeExpr :: ScopeInfo -> Expr -> TCM Expr
--- rescopeExpr scope = withScope_ scope . (concreteToAbstract_ <=< runAbsToCon . preserveInteractionIds . toConcrete)
-
 {-| Evaluate the given expression in the current environment -}
 evalInCurrent :: ComputeMode -> Expr -> TCM Expr
 evalInCurrent cmode e = do
@@ -368,9 +347,11 @@ computeIgnoreAbstract :: ComputeMode -> Bool
 computeIgnoreAbstract DefaultCompute  = False
 computeIgnoreAbstract HeadCompute     = False
 computeIgnoreAbstract IgnoreAbstract  = True
-computeIgnoreAbstract UseShowInstance = True
-  -- UseShowInstance requires the result to be a string literal so respecting
-  -- abstract can only ever break things.
+computeIgnoreAbstract UseShowInstance = False
+  -- WRONG: UseShowInstance requires the result to be a string literal so respecting abstract can
+  -- only ever break things.
+  -- See #7191. Abstract definitions can come with rewrite rules to give nice show behaviour that
+  -- breaks if we don't respect abstract.
 
 computeWrapInput :: ComputeMode -> String -> String
 computeWrapInput UseShowInstance s = "show (" ++ s ++ ")"
@@ -380,7 +361,7 @@ showComputed :: ComputeMode -> Expr -> TCM Doc
 showComputed UseShowInstance e =
   case e of
     A.Lit _ (LitString s) -> pure (text $ T.unpack s)
-    _                     -> ("Not a string:" $$) <$> prettyATop e
+    _                     -> ("Expected applying `show` to the given value to produce a string literal, but got:" $$) <$> prettyATop e
 showComputed _ e = prettyATop e
 
 -- | Modifier for interactive commands,
@@ -405,6 +386,7 @@ outputFormId (OutputForm _ _ _ o) = out o
       IsEmptyType _              -> __IMPOSSIBLE__   -- Should never be used on IsEmpty constraints
       SizeLtSat{}                -> __IMPOSSIBLE__
       FindInstanceOF _ _ _        -> __IMPOSSIBLE__
+      ResolveInstanceOF _        -> __IMPOSSIBLE__
       PTSInstance i _            -> i
       PostponedCheckFunDef{}     -> __IMPOSSIBLE__
       DataSort _ i               -> i
@@ -426,80 +408,85 @@ reifyElimToExpr = \case
     appl s v = A.App defaultAppInfo_ (A.Lit empty (LitString s)) $ fmap unnamed v
 
 instance Reify Constraint where
-    type ReifiesTo Constraint = OutputConstraint Expr Expr
+  type ReifiesTo Constraint = OutputConstraint Expr Expr
 
-    reify (ValueCmp cmp (AsTermsOf t) u v) = CmpInType cmp <$> reify t <*> reify u <*> reify v
-    reify (ValueCmp cmp AsSizes u v) = CmpInType cmp <$> (reify =<< sizeType) <*> reify u <*> reify v
-    reify (ValueCmp cmp AsTypes u v) = CmpTypes cmp <$> reify u <*> reify v
-    reify (ValueCmpOnFace cmp p t u v) = CmpInType cmp <$> (reify =<< ty) <*> reify (lam_o u) <*> reify (lam_o v)
-      where
-        lam_o = I.Lam (setRelevance Irrelevant defaultArgInfo) . NoAbs "_"
-        ty = runNamesT [] $ do
-          p <- open p
-          t <- open t
-          pPi' "o" p (\ o -> t)
-    reify (ElimCmp cmp _ t v es1 es2) =
-      CmpElim cmp <$> reify t <*> mapM reifyElimToExpr es1
-                              <*> mapM reifyElimToExpr es2
-    reify (LevelCmp cmp t t')    = CmpLevels cmp <$> reify t <*> reify t'
-    reify (SortCmp cmp s s')     = CmpSorts cmp <$> reify s <*> reify s'
-    reify (UnquoteTactic tac _ goal) = do
-        tac <- A.App defaultAppInfo_ (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
-        OfType tac <$> reify goal
-    reify (UnBlock m) = do
-        mi <- lookupMetaInstantiation m
-        m' <- reify (MetaV m [])
-        case mi of
-          BlockedConst t -> do
-            e  <- reify t
-            return $ Assign m' e
-          PostponedTypeCheckingProblem cl -> enterClosure cl $ \case
-            CheckExpr cmp e a -> do
-                a  <- reify a
-                return $ TypedAssign m' e a
-            CheckLambda cmp (Arg ai (xs, mt)) body target -> do
-              domType <- maybe (return underscore) reify mt
-              target  <- reify target
-              let mkN (WithHiding h x) = setHiding h $ defaultNamedArg $ A.mkBinder_ x
-                  bs = mkTBind noRange (fmap mkN xs) domType
-                  e  = A.Lam Info.exprNoRange (DomainFull bs) body
-              return $ TypedAssign m' e target
-            CheckArgs _ _ _ args t0 t1 _ -> do
-              t0 <- reify t0
-              t1 <- reify t1
-              return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
-            CheckProjAppToKnownPrincipalArg cmp e _ _ _ t _ _ _ _ -> TypedAssign m' e <$> reify t
-            DoQuoteTerm cmp v t -> do
-              tm <- A.App defaultAppInfo_ (A.QuoteTerm exprNoRange) . defaultNamedArg <$> reify v
-              OfType tm <$> reify t
-          Open{}  -> __IMPOSSIBLE__
-          OpenInstance{}  -> __IMPOSSIBLE__
-          InstV{} -> __IMPOSSIBLE__
-    reify (FindInstance m mcands) = FindInstanceOF
-      <$> reify (MetaV m [])
-      <*> (reify =<< getMetaType m)
-      <*> forM (fromMaybe [] mcands) (\ (Candidate q tm ty _) -> do
-            (,,) <$> reify tm <*> reify tm <*> reify ty)
-    reify (IsEmpty r a) = IsEmptyType <$> reify a
-    reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
-    reify (CheckFunDef i q cs err) = do
-      a <- reify =<< defType <$> getConstInfo q
-      return $ PostponedCheckFunDef q a err
-    reify (HasBiggerSort a) = OfType <$> reify a <*> reify (UnivSort a)
-    reify (HasPTSRule a b) = do
-      (a,(x,b)) <- reify (unDom a,b)
-      return $ PTSInstance a b
-    reify (CheckDataSort q s) = DataSort q <$> reify s
-    reify (CheckLockedVars t _ lk _) = CheckLock <$> reify t <*> reify (unArg lk)
-    reify (CheckMetaInst m) = do
-      t <- jMetaType . mvJudgement <$> lookupLocalMeta m
-      OfType <$> reify (MetaV m []) <*> reify t
-    reify (CheckType t) = JustType <$> reify t
-    reify (UsableAtModality _ _ mod t) = UsableAtMod mod <$> reify t
+  reify (ValueCmp cmp (AsTermsOf t) u v) = CmpInType cmp <$> reify t <*> reify u <*> reify v
+  reify (ValueCmp cmp AsSizes u v) = CmpInType cmp <$> (reify =<< sizeType) <*> reify u <*> reify v
+  reify (ValueCmp cmp AsTypes u v) = CmpTypes cmp <$> reify u <*> reify v
+  reify (ValueCmpOnFace cmp p t u v) = CmpInType cmp <$> (reify =<< ty) <*> reify (lam_o u) <*> reify (lam_o v)
+    where
+      lam_o = I.Lam defaultIrrelevantArgInfo . NoAbs "_"
+      ty = runNamesT [] $ do
+        p <- open p
+        t <- open t
+        pPi' "o" p (\ o -> t)
+  reify (ElimCmp cmp _ t v es1 es2) =
+    CmpElim cmp <$> reify t <*> mapM reifyElimToExpr es1
+                            <*> mapM reifyElimToExpr es2
+  reify (LevelCmp cmp t t')    = CmpLevels cmp <$> reify t <*> reify t'
+  reify (SortCmp cmp s s')     = CmpSorts cmp <$> reify s <*> reify s'
+  reify (UnquoteTactic tac _ goal) = do
+      tac <- A.App defaultAppInfo_ (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
+      OfType tac <$> reify goal
+  reify (UnBlock m) = do
+      mi <- lookupMetaInstantiation m
+      m' <- reify (MetaV m [])
+      case mi of
+        BlockedConst t -> do
+          e  <- reify t
+          return $ Assign m' e
+        PostponedTypeCheckingProblem cl -> enterClosure cl $ \case
+          CheckExpr cmp e a -> do
+              a  <- reify a
+              return $ TypedAssign m' e a
+          CheckLambda cmp (Arg ai (xs, mt)) body target -> do
+            domType <- maybe (return underscore) reify mt
+            target  <- reify target
+            let mkN (WithHiding h x) = setHiding h $ defaultNamedArg $ A.mkBinder_ x
+                bs = mkTBind noRange (fmap mkN xs) domType
+                e  = A.Lam Info.exprNoRange (DomainFull bs) body
+            return $ TypedAssign m' e target
+          CheckArgs _ _ _ args t0 t1 _ -> do
+            t0 <- reify t0
+            t1 <- reify t1
+            return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
+          CheckProjAppToKnownPrincipalArg cmp e _ _ _ _ t _ _ _ _ -> TypedAssign m' e <$> reify t
+          DoQuoteTerm cmp v t -> do
+            tm <- A.App defaultAppInfo_ (A.QuoteTerm exprNoRange) . defaultNamedArg <$> reify v
+            OfType tm <$> reify t
+          DisambiguateConstructor (ConstructorDisambiguationData c0 _cands args t) _cont -> do
+            t <- reify t
+            return $ TypedAssign m' (foldl (A.App empty)  (A.Con $ unambiguous c0) args) t
+        OpenMeta{}  -> __IMPOSSIBLE__
+        InstV{} -> __IMPOSSIBLE__
+  reify (FindInstance m mcands) = FindInstanceOF
+    <$> reify (MetaV m [])
+    <*> (reify =<< getMetaType m)
+    <*> forM (fromMaybe [] mcands) (\ (Candidate q tm ty _) -> do
+          (,,) <$> reify tm <*> reify tm <*> reify ty)
+  reify (ResolveInstanceHead q) = return $ ResolveInstanceOF q
+  reify (IsEmpty r a) = IsEmptyType <$> reify a
+  reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
+  reify (CheckFunDef i q cs err) = do
+    a <- reify =<< defType <$> getConstInfo q
+    return $ PostponedCheckFunDef q a err
+  reify (HasBiggerSort a) = OfType <$> reify a <*> reify (UnivSort a)
+  reify (HasPTSRule a b) = do
+    (a,(x,b)) <- reify (unDom a,b)
+    return $ PTSInstance a b
+  reify (CheckDataSort q s) = DataSort q <$> reify s
+  reify (CheckLockedVars t _ lk _) = CheckLock <$> reify t <*> reify (unArg lk)
+  reify (CheckMetaInst m) = do
+    t <- jMetaType . mvJudgement <$> lookupLocalMeta m
+    OfType <$> reify (MetaV m []) <*> reify t
+  reify (CheckType t) = JustType <$> reify t
+  reify (UsableAtModality _ _ mod t) = UsableAtMod mod <$> reify t
+  {-# SPECIALIZE reify :: Constraint -> TCM (ReifiesTo Constraint) #-}
 
 instance (Pretty a, Pretty b) => PrettyTCM (OutputForm a b) where
   prettyTCM (OutputForm r pids unblock c) =
     prettyRangeConstraint r pids unblock (pretty c)
+  {-# SPECIALIZE prettyTCM :: (Pretty a, Pretty b) => (OutputForm a b) -> TCM Doc #-}
 
 instance (Pretty a, Pretty b) => Pretty (OutputForm a b) where
   pretty (OutputForm r pids unblock c) =
@@ -544,6 +531,8 @@ instance (Pretty a, Pretty b) => Pretty (OutputConstraint a b) where
         [ "Resolve instance argument" <?> (pretty s .: t)
         , nest 2 $ "Candidate:"
         , nest 4 $ vcat [ bin (pretty q) "=" (pretty v) .: t | (q, v, t) <- cs ] ]
+      ResolveInstanceOF q ->
+        "Resolve output type of instance" <?> pretty q
       PTSInstance a b      -> "PTS instance for" <+> pretty (a, b)
       PostponedCheckFunDef q a _err ->
         vcat [ "Check definition of" <+> pretty q <+> ":" <+> pretty a ]
@@ -589,6 +578,7 @@ instance (ToConcrete a, ToConcrete b) => ToConcrete (OutputConstraint a b) where
     toConcrete (FindInstanceOF s t cs) =
       FindInstanceOF <$> toConcrete s <*> toConcrete t
                      <*> mapM (\(q,tm,ty) -> (,,) <$> toConcrete q <*> toConcrete tm <*> toConcrete ty) cs
+    toConcrete (ResolveInstanceOF q) = return $ ResolveInstanceOF q
     toConcrete (PTSInstance a b) = PTSInstance <$> toConcrete a <*> toConcrete b
     toConcrete (DataSort a b)  = DataSort a <$> toConcrete b
     toConcrete (CheckLock a b) = CheckLock <$> toConcrete a <*> toConcrete b
@@ -661,6 +651,7 @@ getConstraintsMentioning norm m = getConstrs instantiateBlockingFull (mentionsMe
         SortCmp cmp a b            -> Nothing
         UnBlock{}                  -> Nothing
         FindInstance{}             -> Nothing
+        ResolveInstanceHead{}      -> Nothing
         IsEmpty r t                -> isMeta (unEl t)
         CheckSizeLtSat t           -> isMeta t
         CheckFunDef{}              -> Nothing
@@ -718,8 +709,8 @@ stripConstraintPids cs = List.sortBy (compare `on` isBlocked) $ map stripPids cs
     interestingPids = Set.unions $ map (allBlockingProblems . constraintUnblocker) cs
     stripPids (PConstr pids unblock c) = PConstr (Set.intersection pids interestingPids) unblock c
 
+{-# SPECIALIZE interactionIdToMetaId :: InteractionId -> TCM MetaId #-}
 -- | Converts an 'InteractionId' to a 'MetaId'.
-
 interactionIdToMetaId :: ReadTCState m => InteractionId -> m MetaId
 interactionIdToMetaId i = do
   h <- currentModuleNameHash
@@ -742,7 +733,8 @@ getConstraints' g f = liftTCM $ do
       withMetaInfo mv $ do
         mi <- interactionIdToMetaId ii
         let m = QuestionMark emptyMetaInfo{ metaNumber = Just mi } ii
-        abstractToConcrete_ $ OutputForm noRange [] alwaysUnblock $ Assign m e
+        let oform = OutputForm noRange [] alwaysUnblock $ Assign m e :: OutputForm Expr Expr
+        abstractToConcrete_ oform
 
 -- | Reify the boundary of an interaction point as something that can be
 -- shown to the user.
@@ -852,10 +844,10 @@ showGoals (ims, hms) = do
 getWarningsAndNonFatalErrors :: TCM WarningsAndNonFatalErrors
 getWarningsAndNonFatalErrors = do
   mws <- getAllWarnings AllWarnings
-  let notMetaWarnings = filter (not . isMetaTCWarning) mws
+  let notMetaWarnings = filter (not . isMetaTCWarning) $ Set.toList mws
   return $ case notMetaWarnings of
     ws@(_:_) -> classifyWarnings ws
-    _ -> emptyWarningsAndNonFatalErrors
+    _ -> empty
 
 -- | Collecting the context of the given meta-variable.
 getResponseContext
@@ -889,8 +881,7 @@ getSolvedInteractionPoints all norm = concat <$> do
             unsol = return []
         case mvInstantiation mv of
           InstV{}                        -> sol (MetaV m $ map Apply args)
-          Open{}                         -> unsol
-          OpenInstance{}                 -> unsol
+          OpenMeta{}                     -> unsol
           BlockedConst{}                 -> unsol
           PostponedTypeCheckingProblem{} -> unsol
 
@@ -948,8 +939,7 @@ typesOfHiddenMetas norm = liftTCM $ do
   implicit is x m =
     case mvInstantiation m of
       M.InstV{} -> __IMPOSSIBLE__
-      M.Open    -> x `notElem` is
-      M.OpenInstance -> x `notElem` is  -- OR: True !?
+      M.OpenMeta _ -> x `notElem` is  -- OR: True in case of InstanceMeta !?
       M.BlockedConst{} -> False
       M.PostponedTypeCheckingProblem{} -> False
 
@@ -967,8 +957,14 @@ metaHelperType norm ii rng s = case words s of
 
     -- Konstantin, 2022-10-23: We don't want to print section parameters in helper type.
     freeVars <- getCurrentModuleFreeVars
-    contextForAbstracting <- drop freeVars . reverse <$> getContext
-    let escapeAbstractedContext = escapeContext impossible (length contextForAbstracting)
+    ctx <- getContext
+    let contextForAbstracting = take (size ctx - freeVars) ctx
+
+    -- Andreas, 2019-10-11: I actually prefer pi-types over ->.
+    let runInPrintingEnvironment = localTC (\e -> e { envPrintDomainFreePi = True, envPrintMetasBare = True })
+                                 . escapeContext impossible (length contextForAbstracting)
+                                 . withoutPrintingGeneralization
+                                 . dontFoldLetBindings
 
     case mapM (isVar . namedArg) args >>= \ xs -> xs <$ guard (all inCxt xs) of
 
@@ -977,44 +973,44 @@ metaHelperType norm ii rng s = case words s of
      -- We simply make exactly the given arguments visible and all other hidden.
      Just xs -> do
       let inXs = hasElem xs
-      let hideButXs dom = setHiding (if inXs $ fst $ unDom dom then NotHidden else Hidden) dom
-      let tel = telFromList . map (fmap (first nameToArgName) . hideButXs) $ contextForAbstracting
+      let hideButXs ce = setHiding (if inXs (ctxEntryName ce) then NotHidden else Hidden) ce
+      let tel = contextToTel . map hideButXs $ contextForAbstracting
       OfType' h <$> do
-        -- Andreas, 2019-10-11: I actually prefer pi-types over ->.
-        localTC (\e -> e { envPrintDomainFreePi = True }) $ escapeAbstractedContext $ withoutPrintingGeneralization $
-          reify $ telePiVisible tel a0
+        runInPrintingEnvironment $ reify $ telePiVisible tel a0
 
-     -- If some arguments are not variables.
+     -- If some arguments are not variables (in this case, @args@ is not empty).
      Nothing -> do
       -- cleanupType relies on with arguments being named 'w',
       -- so we'd better rename any actual 'w's to avoid confusion.
-      let tel = runIdentity . onNamesTel unW . telFromList' nameToArgName $ contextForAbstracting
+      let tel = runIdentity . onNamesTel unW . contextToTel $ contextForAbstracting
       let a = runIdentity . onNames unW $ a0
-      vtys <- mapM (\ a -> fmap (Arg (getArgInfo a) . fmap OtherType) $ inferExpr $ namedArg a) args
+      vtys <- mapM (\ a -> fmap (Arg (getArgInfo a) . fmap OtherType) $ inferExpr $ namedArg a) $
+        List1.fromListSafe __IMPOSSIBLE__ args
       -- Remember the arity of a
       TelV atel _ <- telView a
       let arity = size atel
           (delta1, delta2, _, a', vtys') = splitTelForWith tel a vtys
-      a <- localTC (\e -> e { envPrintDomainFreePi = True, envPrintMetasBare = True }) $ escapeAbstractedContext $ withoutPrintingGeneralization $ do
-        reify =<< cleanupType arity args =<< normalForm norm =<< fst <$> withFunctionType delta1 vtys' delta2 a' []
-      reportSDoc "interaction.helper" 10 $ TP.vcat $
-        let extractOtherType = \case { OtherType a -> a; _ -> __IMPOSSIBLE__ } in
-        let (vs, as)   = unzipWith (fmap extractOtherType . unArg) vtys in
-        let (vs', as') = unzipWith (fmap extractOtherType . unArg) vtys' in
-        [ "generating helper function"
-        , TP.nest 2 $ "tel    = " TP.<+> inTopContext (prettyTCM tel)
-        , TP.nest 2 $ "a      = " TP.<+> prettyTCM a
-        , TP.nest 2 $ "vs     = " TP.<+> prettyTCM vs
-        , TP.nest 2 $ "as     = " TP.<+> prettyTCM as
-        , TP.nest 2 $ "delta1 = " TP.<+> inTopContext (prettyTCM delta1)
-        , TP.nest 2 $ "delta2 = " TP.<+> inTopContext (addContext delta1 $ prettyTCM delta2)
-        , TP.nest 2 $ "a'     = " TP.<+> inTopContext (addContext delta1 $ addContext delta2 $ prettyTCM a')
-        , TP.nest 2 $ "as'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM as')
-        , TP.nest 2 $ "vs'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM vs')
-        ]
+      a <- runInPrintingEnvironment $ do
+        reify =<< cleanupType arity args =<< normalForm norm =<< fst <$> withFunctionType delta1 vtys' delta2 a' empty
+      reportSDoc "interaction.helper" 10 do
+        let extractOtherType = \case { OtherType a -> a; _ -> __IMPOSSIBLE__ }
+        let (vs, as)   = List1.unzipWith (fmap extractOtherType . unArg) vtys
+        let (vs', as') = List1.unzipWith (fmap extractOtherType . unArg) vtys'
+        TP.vcat
+          [ "generating helper function"
+          , TP.nest 2 $ "tel    = " TP.<+> inTopContext (prettyTCM tel)
+          , TP.nest 2 $ "a      = " TP.<+> prettyTCM a
+          , TP.nest 2 $ "vs     = " TP.<+> prettyTCM vs
+          , TP.nest 2 $ "as     = " TP.<+> prettyTCM as
+          , TP.nest 2 $ "delta1 = " TP.<+> inTopContext (prettyTCM delta1)
+          , TP.nest 2 $ "delta2 = " TP.<+> inTopContext (addContext delta1 $ prettyTCM delta2)
+          , TP.nest 2 $ "a'     = " TP.<+> inTopContext (addContext delta1 $ addContext delta2 $ prettyTCM a')
+          , TP.nest 2 $ "as'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM as')
+          , TP.nest 2 $ "vs'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM vs')
+          ]
       return $ OfType' h a
   where
-    failure = typeError $ GenericError $ "Expected an argument of the form f e1 e2 .. en"
+    failure = interactionError ExpectedApplication
     ensureName f = do
       ce <- parseExpr rng f
       flip (caseMaybe $ isName ce) (\ _ -> return ()) $ do
@@ -1114,7 +1110,7 @@ contextOfMeta ii norm = withInteractionId ii $ do
 
   where
     mkVar :: ContextEntry -> TCM (Maybe ResponseContextEntry)
-    mkVar Dom{ domInfo = ai, unDom = (name, t) } = do
+    mkVar (CtxVar name Dom{ domInfo = ai, unDom = t }) = do
       if shouldHide ai name then return Nothing else Just <$> do
         let n = nameConcrete name
         x  <- abstractToConcrete_ name
@@ -1174,7 +1170,7 @@ introTactic pmLambda ii = do
         TelV tel' t <- telViewUpTo' (-1) notVisible t
         -- if we cannot introduce a constructor, we try a lambda
         let fallback = do
-              cubical <- isJust . optCubical <$> pragmaOptions
+              cubical <- isJust <$> cubicalOption
               TelV tel _ <- (if cubical then telViewPath else telView) t
               reportSDoc "interaction.intro" 20 $ TP.sep
                 [ "introTactic/fallback"
@@ -1202,11 +1198,12 @@ introTactic pmLambda ii = do
     conName [p] = [ c | I.ConP c _ _ <- [namedArg p] ]
     conName _   = __IMPOSSIBLE__
 
-    showUnambiguousConName amb v =
-       render . pretty <$> runAbsToCon (lookupQName amb $ I.conName v)
+    showUnambiguousConName :: AllowAmbiguousNames -> ConHead -> TCM String
+    showUnambiguousConName amb c = render . pretty <$> do
+      abstractToConcreteQName amb $ I.conName c
 
     showTCM :: PrettyTCM a => a -> TCM String
-    showTCM v = render <$> prettyTCM v
+    showTCM = render <.> prettyTCM
 
     introFun :: ListTel -> TCM [String]
     introFun tel = addContext tel' $ do
@@ -1236,11 +1233,11 @@ introTactic pmLambda ii = do
     introData :: AllowAmbiguousNames -> I.Type -> TCM [String]
     introData amb t = do
       let tel  = telFromList [defaultDom ("_", t)]
-          pat  = [defaultArg $ unnamed $ debruijnNamedVar "c" 0]
+          pat  = [defaultArg $ unnamed $ deBruijnNamedVar "c" 0]
       -- Gallais, 2023-08-24: #6787 we need to locally ignore the
       -- --without-K or --cubical-compatible options to figure out
       -- that refl is a valid constructor for refl ≡ refl.
-      cubical <- isJust . optCubical <$> pragmaOptions
+      cubical <- isJust <$> cubicalOption
       r <- (if cubical then id else
             locallyTCState (stPragmaOptions . lensOptWithoutK) (const (Value False)))
            $ splitLast CoInductive tel pat
@@ -1269,7 +1266,9 @@ introTactic pmLambda ii = do
 --   Sets up current module, scope, and context.
 atTopLevel :: TCM a -> TCM a
 atTopLevel m = inConcreteMode $ do
-  let err = typeError $ GenericError "The file has not been loaded yet."
+  let err = __IMPOSSIBLE__
+    -- Andreas, 2024-08-03: cannot trigger this error:
+    -- let err = genericError "The file has not been loaded yet."
   caseMaybeM (useTC stCurrentModule) err $ \(current, topCurrent) -> do
     caseMaybeM (getVisitedModule topCurrent) __IMPOSSIBLE__ $ \ mi -> do
       let scope = iInsideScope $ miInterface mi
@@ -1320,8 +1319,7 @@ atTopLevel m = inConcreteMode $ do
 parseName :: Range -> String -> TCM C.QName
 parseName r s = do
   e <- parseExpr r s
-  let failure = typeError $ GenericError $ "Not an identifier: " ++ show e ++ "."
-  maybe failure return $ isQName e
+  maybe (interactionError $ ExpectedIdentifier e) return $ isQName e
 
 -- | Check whether an expression is a (qualified) identifier.
 isQName :: C.Expr -> Maybe C.QName
@@ -1371,12 +1369,10 @@ getRecordContents
               --   context extension,
               --   names paired up with corresponding types.
 getRecordContents norm ce = do
-  e <- toAbstract ce
-  (_, t) <- inferExpr e
-  let notRecordType = typeError $ ShouldBeRecordType t
-  (q, vs, defn) <- fromMaybeM notRecordType $ isRecordType t
-  case defn of
-    Record{ recFields = fs, recTel = rtel } -> do
+  (_, t) <- inferExpr =<< toAbstract ce
+  isRecordType t >>= \case
+    Nothing -> typeError $ ShouldBeRecordType t
+    Just (q, vs, RecordData{ _recFields = fs, _recTel = rtel }) -> do
       let xs   = map (nameConcrete . qnameName . unDom) fs
           tel  = apply rtel vs
           doms = flattenTel tel
@@ -1391,7 +1387,6 @@ getRecordContents norm ce = do
         ]
       ts <- mapM (normalForm norm . unDom) doms
       return ([], tel, zip xs ts)
-    _ -> __IMPOSSIBLE__
 
 -- | Returns the contents of the given module.
 
