@@ -1,7 +1,6 @@
 {-# OPTIONS_GHC -Wunused-imports #-}
 
-{-# LANGUAGE GADTs #-}
-{-# LANGUAGE PatternSynonyms #-}
+{-# LANGUAGE DataKinds #-}
 
 {-| The parser doesn't know about operators and parses everything as normal
     function application. This module contains the functions that parses the
@@ -14,17 +13,16 @@
 
 module Agda.Syntax.Concrete.Operators
     ( parseApplication
-    , parseModuleApplication
+    , parseArguments
     , parseLHS
     , parsePattern
     , parsePatternSyn
     ) where
 
-import Control.Applicative ( Alternative((<|>)))
+import Control.Applicative ( Alternative( (<|>) ) )
 import Control.Monad.Except (throwError)
 
 import Data.Either (partitionEithers)
-import qualified Data.Foldable as Fold
 import qualified Data.Function
 import qualified Data.List as List
 import Data.Maybe
@@ -48,11 +46,11 @@ import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.State (getScope)
 
-import Agda.Utils.Function (applyWhen)
+import Agda.Utils.Function (applyWhen, applyWhenJust)
 import Agda.Utils.Either
 import Agda.Syntax.Common.Pretty
 import Agda.Utils.List
-import Agda.Utils.List1 (List1, pattern (:|))
+import Agda.Utils.List1 (List1, pattern (:|), (<|))
 import Agda.Utils.List2 (List2, pattern List2)
 import qualified Agda.Utils.List1 as List1
 import qualified Agda.Utils.List2 as List2
@@ -90,10 +88,6 @@ data InternalParsers e = InternalParsers
   , pAtom   :: Parser e e
   }
 
--- | Expression kinds: Expressions or patterns.
-data ExprKind = IsExpr | IsPattern
-  deriving (Eq, Show)
-
 -- | The data returned by 'buildParsers'.
 
 data Parsers e = Parsers
@@ -130,6 +124,8 @@ buildParsers
   :: forall e. IsExpr e
   => ExprKind
      -- ^ Should expressions or patterns be parsed?
+  -> Maybe QName
+     -- ^ Are we trying to parse the lhs of the function given here?
   -> [QName]
      -- ^ This list must include every name part in the
      -- expression/pattern to be parsed (excluding name parts inside
@@ -139,10 +135,11 @@ buildParsers
      -- grammar if all of the notation's name parts are present in
      -- the list of names.
   -> ScopeM (Parsers e)
-buildParsers kind exprNames = do
+buildParsers kind top exprNames0 = do
+    let exprNames = applyWhenJust top (:) exprNames0
     flat         <- flattenScope (qualifierModules exprNames) <$>
                       getScope
-    (names, ops0) <- localNames flat
+    (names, ops0) <- localNames kind top flat
     let ops | kind == IsPattern = filter (not . isLambdaNotation) ops0
             | otherwise         = ops0
 
@@ -329,7 +326,7 @@ buildParsers kind exprNames = do
 
     let g = Data.Function.fix $ \p -> InternalParsers
               { pTop    = memoise TopK $
-                          Fold.asum $
+                          Agda.Utils.List.asum $
                             foldr (\(l, ns) higher ->
                                        mkP (Right l) parseSections
                                            (pTop p) ns higher True) (pApp p)
@@ -340,7 +337,7 @@ buildParsers kind exprNames = do
               , pApp    = memoise AppK $ appP (pNonfix p) (pArgs p)
               , pArgs   = argsP (pNonfix p)
               , pNonfix = memoise NonfixK $
-                          Fold.asum $
+                          Agda.Utils.List.asum $
                             pAtom p :
                             map (\sect ->
                               let n = sectNotation sect
@@ -365,8 +362,8 @@ buildParsers kind exprNames = do
               }
 
     -- Andreas, 2020-06-03 #4712
-    -- Note: needs Agda to be compiled with DEBUG to print the grammar.
-    reportSDoc "scope.grammar" 10 $ return $
+    -- Note: needs Agda to be compiled with DEBUG_PARSING to print the grammar.
+    reportSDoc "scope.grammar" 20 $ return $
       "Operator grammar:" $$ nest 2 (grammar (pTop g))
 
     return $ Parsers
@@ -405,36 +402,66 @@ buildParsers kind exprNames = do
             -> Parser e e
         mkP key parseSections p0 ops higher includeHigher =
             memoise (NodeK key) $
-              Fold.asum $
+              Agda.Utils.List.asum $
                 applyWhen includeHigher (higher :) $
                 catMaybes [nonAssoc, preRights, postLefts]
-            where
-            choice :: forall k.
-                      NK k -> [NotationSection] ->
-                      Parser e (OperatorType k e)
-            choice k =
-              Fold.asum .
-              map (\sect ->
+          where
+            -- Andreas, 2025-02-27
+            -- Break up the choice function into its three cases,
+            -- so that matching on @k@ does not have to be performed
+            -- inside the mapped function @(\ sect -> ...)@.
+            --
+            -- choice :: forall k.
+            --           NK k -> [NotationSection] ->
+            --           Parser e (OperatorType k e)
+            -- choice k =
+            --   Agda.Utils.List.asum .
+            --   map (\sect ->
+            --     let n = sectNotation sect
+
+            --         inner :: forall k.
+            --                  NK k -> Parser e (OperatorType k e)
+            --         inner = opP parseSections p0 n
+            --     in
+            --     case k of
+            --       In   -> inner In
+
+            --       Pre  -> if isinfix n || ispostfix n
+            --               then flip ($) <$> placeholder Beginning
+            --                             <*> inner In
+            --               else inner Pre
+
+            --       Post -> if isinfix n || isprefix n
+            --               then flip <$> inner In
+            --                         <*> placeholder End
+            --               else inner Post
+
+            --       Non  -> __IMPOSSIBLE__)
+
+            choiceIn :: [NotationSection] -> Parser e (OperatorType 'InfixNotation e)
+            choiceIn =
+              Agda.Utils.List.asum .
+              map \ sect -> opP parseSections p0 (sectNotation sect) In
+
+            choicePre :: [NotationSection] -> Parser e (OperatorType 'PrefixNotation e)
+            choicePre =
+              Agda.Utils.List.asum .
+              map \ sect -> do
                 let n = sectNotation sect
+                if   isinfix n || ispostfix n
+                then flip ($) <$> placeholder Beginning
+                              <*> opP parseSections p0 n In
+                else opP parseSections p0 n Pre
 
-                    inner :: forall k.
-                             NK k -> Parser e (OperatorType k e)
-                    inner = opP parseSections p0 n
-                in
-                case k of
-                  In   -> inner In
-
-                  Pre  -> if isinfix n || ispostfix n
-                          then flip ($) <$> placeholder Beginning
-                                        <*> inner In
-                          else inner Pre
-
-                  Post -> if isinfix n || isprefix n
-                          then flip <$> inner In
-                                    <*> placeholder End
-                          else inner Post
-
-                  Non  -> __IMPOSSIBLE__)
+            choicePost :: [NotationSection] -> Parser e (OperatorType 'PostfixNotation e)
+            choicePost =
+              Agda.Utils.List.asum .
+              map \ sect -> do
+                let n = sectNotation sect
+                if isinfix n || isprefix n
+                then flip <$> opP parseSections p0 n In
+                          <*> placeholder End
+                else opP parseSections p0 n Post
 
             nonAssoc :: Maybe (Parser e e)
             nonAssoc = case filter (isInfix NonAssoc) ops of
@@ -442,7 +469,7 @@ buildParsers kind exprNames = do
               ops -> Just $
                 (\x f y -> f (noPlaceholder x) (noPlaceholder y))
                   <$> higher
-                  <*> choice In ops
+                  <*> choiceIn ops
                   <*> higher
 
             or p1 []   p2 []   = Nothing
@@ -452,10 +479,10 @@ buildParsers kind exprNames = do
 
             preRight :: Maybe (Parser e (MaybePlaceholder e -> e))
             preRight =
-              or (choice Pre)
+              or choicePre
                  (filter isPrefix ops)
                  (\ops -> flip ($) <$> (noPlaceholder <$> higher)
-                                   <*> choice In ops)
+                                   <*> choiceIn ops)
                  (filter (isInfix RightAssoc) ops)
 
             preRights :: Maybe (Parser e e)
@@ -467,9 +494,9 @@ buildParsers kind exprNames = do
 
             postLeft :: Maybe (Parser e (MaybePlaceholder e -> e))
             postLeft =
-              or (choice Post)
+              or choicePost
                  (filter isPostfix ops)
-                 (\ops -> flip <$> choice In ops
+                 (\ops -> flip <$> choiceIn ops
                                <*> (noPlaceholder <$> higher))
                  (filter (isInfix LeftAssoc) ops)
 
@@ -486,31 +513,39 @@ buildParsers kind exprNames = do
 -- * Parse functions
 ---------------------------------------------------------------------------
 
--- | Returns the list of possible parses.
+-- | Parses all 'RawAppP' in the given pattern using the given parser.
+--   Returns the list of possible parses.
+--
+--   Naturally, does not recurse into 'DotP' as this contains no pattern.
+--
+--   Returns the empty list if the given parser does so
+--   or if a 'HiddenP' or 'InstanceP' is encountered.
 parsePat
   :: ([Pattern] -> [Pattern]) -- ^ Turns a 'RawAppP' into possible parses.
   -> Pattern                  -- ^ Pattern possibly containing 'RawAppP's.
   -> [Pattern]                -- ^ Possible parses, not containing 'RawAppP's.
-parsePat prs = \case
+parsePat parse = loop
+  where
+  loop = \case
     AppP p (Arg info q) ->
-        fullParen' <$> (AppP <$> parsePat prs p <*> (Arg info <$> traverse (parsePat prs) q))
-    RawAppP _ ps     -> fullParen' <$> (parsePat prs =<< prs (List2.toList ps))
-    OpAppP r d ns ps -> fullParen' . OpAppP r d ns <$> (mapM . traverse . traverse) (parsePat prs) ps
+        fullParen' <$> (AppP <$> loop p <*> (Arg info <$> traverse loop q))
+    RawAppP _ ps     -> fullParen' <$> (loop =<< parse (List2.toList ps))
+    OpAppP r d ns ps -> fullParen' . OpAppP r d ns <$> (mapM . traverse . traverse) loop ps
     HiddenP _ _      -> fail "bad hidden argument"
     InstanceP _ _    -> fail "bad instance argument"
-    AsP r x p        -> AsP r x <$> parsePat prs p
+    AsP r x p        -> AsP r x <$> loop p
     p@DotP{}         -> return p
-    ParenP r p       -> fullParen' <$> parsePat prs p
+    ParenP r p       -> fullParen' <$> loop p
     p@WildP{}        -> return p
     p@AbsurdP{}      -> return p
     p@LitP{}         -> return p
     p@QuoteP{}       -> return p
     p@IdentP{}       -> return p
-    RecP r fs        -> RecP r <$> mapM (traverse (parsePat prs)) fs
+    RecP r fs        -> RecP r <$> mapM (traverse loop) fs
     p@EqualP{}       -> return p -- Andrea: cargo culted from DotP
     EllipsisP r mp   -> caseMaybe mp (fail "bad ellipsis") $ \p ->
-                          EllipsisP r . Just <$> parsePat prs p
-    WithP r p        -> WithP r <$> parsePat prs p
+                          EllipsisP r . Just <$> loop p
+    WithP r p        -> WithP r <$> loop p
 
 
 {- Implement parsing of copattern left hand sides, e.g.
@@ -561,8 +596,13 @@ instance Pretty ParseLHS where
 
 -- | Parses a left-hand side, workhorse for 'parseLHS'.
 --
-parseLHS'
-  :: LHSOrPatSyn
+parseLHS' ::
+     DisplayLHS
+       -- ^ Are we parsing a 'DisplayPragma'?
+       --   Then defined names are recognized as constructors.
+       --
+       --   In this case, 'LHSOrPatSyn' is 'IsLHS' and 'Maybe QName' is 'Just'.
+  -> LHSOrPatSyn
        -- ^ Are we trying to parse a lhs or a pattern synonym?
        --   For error reporting only!
   -> Maybe QName
@@ -571,23 +611,23 @@ parseLHS'
   -> Pattern
        -- ^ Thing to parse.
   -> ScopeM (ParseLHS, [NotationSection])
-       -- ^ The returned list contains all operators/notations/sections that
-       -- were used to generate the grammar.
+       -- ^ The returned list contains all operators\/notations\/sections that
+       --   were used to generate the grammar.
 
-parseLHS' IsLHS (Just qn) WildP{} =
+parseLHS' NoDisplayLHS IsLHS (Just qn) WildP{} =
     return (ParseLHS qn $ LHSHead qn [], [])
 
-parseLHS' lhsOrPatSyn top p = do
+parseLHS' displayLhs lhsOrPatSyn top p = do
 
     -- Build parser.
-    patP <- buildParsers IsPattern (patternQNames p)
+    patP <- buildParsers IsPattern top (patternQNames p)
 
     -- Run parser, forcing result.
     let ps   = let result = parsePat (parser patP) p
                in  foldr seq () result `seq` result
 
     -- Classify parse results.
-    let cons = getNames (someKindsOfNames [ConName, CoConName, PatternSynName])
+    let cons = getNames (someKindsOfNames $ applyWhen displayLhs (defNameKinds ++) conLikeNameKinds)
                         (flattenedScope patP)
     let flds = getNames (someKindsOfNames [FldName])
                         (flattenedScope patP)
@@ -598,15 +638,15 @@ parseLHS' lhsOrPatSyn top p = do
       [ "Possible parses for lhs:" ] ++ map (nest 2 . pretty . snd) results
     case results of
         -- Unique result.
-        [(_,lhs)] -> do reportS "scope.operators" 50 $ "Parsed lhs:" <+> pretty lhs
-                        return (lhs, operators patP)
+        [(_,lhs)] -> (lhs, operators patP) <$ do
+                       reportS "scope.operators" 50 $ "Parsed lhs:" <+> pretty lhs
         -- No result.
-        []        -> typeError $ OperatorInformation (operators patP)
-                               $ NoParseForLHS lhsOrPatSyn (catMaybes errs) p
+        []        -> typeError $ OperatorInformation (operators patP) $
+                       NoParseForLHS lhsOrPatSyn (catMaybes errs) p
         -- Ambiguous result.
-        rs        -> typeError $ OperatorInformation (operators patP)
-                               $ AmbiguousParseForLHS lhsOrPatSyn p $
-                       map (fullParen . fst) rs
+        r0:r1:rs  -> typeError $ OperatorInformation (operators patP) $
+                       AmbiguousParseForLHS lhsOrPatSyn p $
+                         fmap (fullParen . fst) $ List2 r0 r1 rs
     where
         getNames kinds flat =
           map (notaName . List1.head) $ getDefinedNames kinds flat
@@ -690,9 +730,16 @@ classifyPattern conf p =
 
 
 -- | Parses a left-hand side, and makes sure that it defined the expected name.
-parseLHS :: QName -> Pattern -> ScopeM LHSCore
-parseLHS top p = billToParser IsPattern $ do
-  (res, ops) <- parseLHS' IsLHS (Just top) p
+parseLHS ::
+     DisplayLHS
+       -- ^ Are we parsing a 'DisplayPragma'?
+  -> QName
+       -- ^ Name of the definition.
+  -> Pattern
+       -- ^ Full left hand side.
+  -> ScopeM LHSCore
+parseLHS displayLhs top p = billToParser IsPattern $ do
+  (res, ops) <- parseLHS' displayLhs IsLHS (Just top) p
   case res of
     ParseLHS f lhs -> return lhs
     _ -> typeError $ OperatorInformation ops
@@ -707,7 +754,7 @@ parsePatternSyn = parsePatternOrSyn IsPatSyn
 
 parsePatternOrSyn :: LHSOrPatSyn -> Pattern -> ScopeM Pattern
 parsePatternOrSyn lhsOrPatSyn p = billToParser IsPattern $ do
-  (res, ops) <- parseLHS' lhsOrPatSyn Nothing p
+  (res, ops) <- parseLHS' NoDisplayLHS lhsOrPatSyn Nothing p
   case res of
     ParsePattern p -> return p
     _ -> typeError $ OperatorInformation ops
@@ -724,14 +771,34 @@ validConPattern
 validConPattern cons = loop
   where
   loop p = case appView p of
-      WithP _ p :| [] -> loop p
-      _ :| []         -> ok
-      IdentP _ x :| ps
-        | cons x      -> mapM_ loop ps
-        | otherwise   -> failure
-      QuoteP _ :| [_] -> ok
-      DotP _ e :| ps  -> mapM_ loop ps
-      _               -> failure
+
+      -- Eliminated by appView:
+      AppP{}      :| _   -> __IMPOSSIBLE__
+      OpAppP{}    :| _   -> __IMPOSSIBLE__
+      ParenP{}    :| _   -> __IMPOSSIBLE__
+      RawAppP{}   :| _   -> __IMPOSSIBLE__
+      HiddenP{}   :| _   -> __IMPOSSIBLE__
+      InstanceP{} :| _   -> __IMPOSSIBLE__
+
+      -- Hopeful cases:
+      WithP _ p   :| []  -> loop p
+      _           :| []  -> ok
+      IdentP _ x  :| ps
+        | cons x         -> mapM_ loop ps
+        | otherwise      -> failure
+      QuoteP _    :| [_] -> ok
+      DotP _ _ _  :| ps  -> mapM_ loop ps
+
+      -- Failures:
+      AbsurdP{}   :| _:_ -> failure
+      AsP{}       :| _:_ -> failure
+      EllipsisP{} :| _:_ -> failure
+      EqualP{}    :| _:_ -> failure
+      LitP{}      :| _:_ -> failure
+      QuoteP{}    :| _:_ -> failure
+      RecP{}      :| _:_ -> failure
+      WildP{}     :| _:_ -> failure
+      WithP{}     :| _:_ -> failure
     where
     ok      = return ()
     failure = throwError $ Just p
@@ -743,7 +810,7 @@ appView = loop []
   where
   loop acc = \case
     AppP p a         -> loop (namedArg a : acc) p
-    OpAppP _ op _ ps -> (IdentP True op :| fmap namedArg ps)
+    OpAppP _ op _ ps -> (IdentP True op <| fmap namedArg ps)
                           `List1.appendList`
                         reverse acc
     ParenP _ p       -> loop acc p
@@ -776,7 +843,7 @@ parseApplication :: List2 Expr -> ScopeM Expr
 parseApplication es  = billToParser IsExpr $ do
     let es0 = List2.toList es
     -- Build the parser
-    p <- buildParsers IsExpr [ q | Ident q <- es0 ]
+    p <- buildParsers IsExpr Nothing [ q | Ident q <- es0 ]
 
     -- Parse
     let result = parser p es0
@@ -791,41 +858,34 @@ parseApplication es  = billToParser IsExpr $ do
                          $ AmbiguousParseForApplication es
                          $ fmap fullParen (e :| es')
 
-parseModuleIdentifier :: Expr -> ScopeM QName
-parseModuleIdentifier (Ident m) = return m
-parseModuleIdentifier e = typeError $ NotAModuleExpr e
+-- | Parse the arguments of a raw application with known head.
+--
+parseArguments ::
+     Expr                   -- ^ Head
+  -> [Expr]                 -- ^ Raw arguments
+  -> ScopeM [NamedArg Expr] -- ^ Operator-parsed arguments
+parseArguments hd = \case
+  [] -> return []
+  es@(e1 : rest) -> billToParser IsExpr $ do
 
-parseRawModuleApplication :: List2 Expr -> ScopeM (QName, [NamedArg Expr])
-parseRawModuleApplication es@(List2 e e2 rest) = billToParser IsExpr $ do
-    let es_args = e2:rest
-    m <- parseModuleIdentifier e
+    -- Form the raw application for error reporting
+    let es2 = List2 hd e1 rest
 
     -- Build the arguments parser
-    p <- buildParsers IsExpr [ q | Ident q <- es_args ]
+    p <- buildParsers IsExpr Nothing [ q | Ident q <- es ]
 
     -- Parse
     -- TODO: not sure about forcing
-    case {-force $-} argsParser p es_args of
-        [as] -> return (m, as)
+    case {-force $-} argsParser p es of
+        [as] -> return as
         []   -> typeError $ OperatorInformation (operators p)
-                          $ NoParseForApplication es
+                          $ NoParseForApplication es2
         as : ass -> do
-          let f = fullParen . foldl (App noRange) (Ident m)
+          let f = fullParen . foldl (App noRange) hd
           typeError $ OperatorInformation (operators p)
-                    $ AmbiguousParseForApplication es
+                    $ AmbiguousParseForApplication es2
                     $ fmap f (as :| ass)
 
--- | Parse an expression into a module application
---   (an identifier plus a list of arguments).
-parseModuleApplication :: Expr -> ScopeM (QName, [NamedArg Expr])
-parseModuleApplication (RawApp _ es) = parseRawModuleApplication es
-parseModuleApplication (App r e1 e2) = do -- TODO: do we need this case?
-    (m, args) <- parseModuleApplication e1
-    return (m, args ++ [e2])
-parseModuleApplication e = do
-    m <- parseModuleIdentifier e
-    return (m, [])
-
 ---------------------------------------------------------------------------
 -- * Inserting parenthesis
 ---------------------------------------------------------------------------
