@@ -60,16 +60,14 @@
 module Agda.TypeChecking.Free.Lazy where
 
 import Control.Applicative hiding (empty)
-import Control.Monad        ( guard )
 import Control.Monad.Reader ( MonadReader(..), asks, ReaderT, Reader, runReader )
 
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import Data.HashSet (HashSet)
 import qualified Data.HashSet as HashSet
-import Data.Semigroup ( Semigroup, (<>) )
-import qualified Data.Set as Set
 import Data.Set (Set)
+import qualified Data.Set as Set
 
 
 import Agda.Syntax.Common
@@ -77,6 +75,7 @@ import Agda.Syntax.Internal
 
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.List1 (List1)
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Semigroup
@@ -470,6 +469,12 @@ underModality = local . mapModality . composeModality . getModality
 underRelevance :: (MonadReader r m, LensRelevance r, LensRelevance o) => o -> m z -> m z
 underRelevance = local . mapRelevance . composeRelevance . getRelevance
 
+-- | In the given computation the 'Quantity' is locally scaled using
+-- the 'Quantity' of the first argument.
+underQuantity ::
+  (MonadReader r m, LensQuantity r, LensQuantity o) => o -> m a -> m a
+underQuantity = local . mapQuantity . composeQuantity . getQuantity
+
 -- | Changing the 'FlexRig' context.
 underFlexRig :: (MonadReader r m, LensFlexRig r a, Semigroup a, LensFlexRig o a) => o -> m z -> m z
 underFlexRig = local . over lensFlexRig . composeFlexRig . view lensFlexRig
@@ -533,7 +538,7 @@ instance Free Term where
     Sort s       -> freeVars' s
     Level l      -> freeVars' l
     MetaV m ts   -> underFlexRig (Flexible $ singleton m) $ freeVars' ts
-    DontCare mt  -> underModality (Modality Irrelevant unitQuantity unitCohesion) $ freeVars' mt
+    DontCare mt  -> underModality (Modality irrelevant unitQuantity unitCohesion unitPolarity) $ freeVars' mt
     Dummy{}      -> mempty
 
 instance Free t => Free (Type' t) where
@@ -566,9 +571,10 @@ instance Free Level where
 instance Free t => Free (PlusLevel' t) where
   freeVars' (Plus _ l)    = freeVars' l
 
-instance Free t => Free [t]            where
-instance Free t => Free (Maybe t)      where
-instance Free t => Free (WithHiding t) where
+instance Free t => Free [t]
+instance Free t => Free (List1 t)
+instance Free t => Free (Maybe t)
+instance Free t => Free (WithHiding t)
 instance Free t => Free (Named nm t)
 
 instance (Free t, Free u) => Free (t, u) where
