@@ -4,8 +4,6 @@ module Agda.TypeChecking.Monad.Signature where
 
 import Prelude hiding (null)
 
-import qualified Control.Monad.Fail as Fail
-
 import Control.Arrow                 ( first, second )
 import Control.Monad.Except          ( ExceptT )
 import Control.Monad.State           ( StateT  )
@@ -51,15 +49,14 @@ import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage.SplitTree
+import {-# SOURCE #-} Agda.TypeChecking.InstanceArguments
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
 import {-# SOURCE #-} Agda.TypeChecking.Pretty
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
 import {-# SOURCE #-} Agda.TypeChecking.Reduce
 import {-# SOURCE #-} Agda.TypeChecking.Opacity
-
-import {-# SOURCE #-} Agda.Compiler.Treeless.Erase
-import {-# SOURCE #-} Agda.Compiler.Builtin
+import {-# SOURCE #-} Agda.TypeChecking.Telescope
 
 import Agda.Utils.CallStack.Base
 import Agda.Utils.Either
@@ -181,17 +178,24 @@ addConstant q d = do
   i <- currentOrFreshMutualBlock
   setMutualBlock i q
   where
-    new +++ old = new { defDisplay = defDisplay new ++ defDisplay old
-                      , defInstance = defInstance new `mplus` defInstance old }
+    new +++ old = new { defDisplay        = defDisplay new ++ defDisplay old
+                      , defInstance       = defInstance new `mplus` defInstance old
+                      , defArgOccurrences = if null (defArgOccurrences new)
+                                              then defArgOccurrences old
+                                              else defArgOccurrences new
+                      , defPolarity       = if null (defPolarity new)
+                                              then defPolarity old
+                                              else defPolarity new
+                      }
 
 -- | A combination of 'addConstant' and 'defaultDefn'. The 'Language'
 -- does not need to be supplied.
 
 addConstant' ::
-  QName -> ArgInfo -> QName -> Type -> Defn -> TCM ()
-addConstant' q info x t def = do
+  QName -> ArgInfo -> Type -> Defn -> TCM ()
+addConstant' q info t def = do
   lang <- getLanguage
-  addConstant q $ defaultDefn info x t lang def
+  addConstant q $ defaultDefn info q t lang def
 
 -- | Set termination info of a defined function symbol.
 setTerminates :: MonadTCState m => QName -> Bool -> m ()
@@ -237,37 +241,20 @@ mkPragma s = CompilerPragma <$> getCurrentRange <*> pure s
 
 -- | Add a compiler pragma `{-\# COMPILE <backend> <name> <text> \#-}`
 addPragma :: BackendName -> QName -> String -> TCM ()
-addPragma b q s = ifM erased
-  {- then -} (warning $ PragmaCompileErased b q)
-  {- else -} $ do
+addPragma b q s = do
     pragma <- mkPragma s
     modifySignature $ updateDefinition q $ addCompilerPragma b pragma
 
-  where
-
-  erased :: TCM Bool
-  erased = do
-    def <- theDef <$> getConstInfo q
-    case def of
-      -- If we have a defined symbol, we check whether it is erasable
-      Function{} ->
-        locallyTC      eActiveBackendName (const $ Just b) $
-        locallyTCState stBackends         (const $ builtinBackends) $
-        isErasable q
-     -- Otherwise (Axiom, Datatype, Record type, etc.) we keep it
-      _ -> pure False
-
 getUniqueCompilerPragma :: BackendName -> QName -> TCM (Maybe CompilerPragma)
 getUniqueCompilerPragma backend q = do
   ps <- defCompilerPragmas backend <$> getConstInfo q
   case ps of
     []  -> return Nothing
     [p] -> return $ Just p
-    (_:p1:_) ->
-      setCurrentRange p1 $
-            genericDocError =<< do
-                  hang (text ("Conflicting " ++ backend ++ " pragmas for") <+> pretty q <+> "at") 2 $
-                       vcat [ "-" <+> pretty (getRange p) | p <- ps ]
+    _:p1:_ -> setCurrentRange p1 do
+      typeError . CustomBackendError backend =<< do
+        hang (hsep [ "Conflicting", pretty backend, "pragmas for", prettyTCM q, "at" ]) 2 $
+          vcat [ "-" <+> pretty (getRange p) | p <- ps ]
 
 setFunctionFlag :: FunctionFlag -> Bool -> QName -> TCM ()
 setFunctionFlag flag val q = modifyGlobalDefinition q $ set (lensTheDef . funFlag flag) val
@@ -281,13 +268,17 @@ markInline b = setFunctionFlag FunInline b
 markInjective :: QName -> TCM ()
 markInjective q = modifyGlobalDefinition q $ \def -> def { defInjective = True }
 
+markFirstOrder :: QName -> TCM ()
+markFirstOrder = setFunctionFlag FunFirstOrder True
+
 unionSignatures :: [Signature] -> Signature
 unionSignatures ss = foldr unionSignature emptySignature ss
   where
-    unionSignature (Sig a b c) (Sig a' b' c') =
+    unionSignature (Sig a b c d) (Sig a' b' c' d') =
       Sig (Map.union a a')
-          (HMap.union b b')              -- definitions are unique (in at most one module)
-          (HMap.unionWith mappend c c')  -- rewrite rules are accumulated
+          (HMap.union b b')             -- definitions are unique (in at most one module)
+          (HMap.unionWith mappend c c') -- rewrite rules are accumulated
+          (d <> d')                     -- instances are accumulated
 
 -- | Add a section to the signature.
 --
@@ -445,30 +436,40 @@ applySection new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd } = d
     -- and if a constructor is copied its datatype needs to be.
     closeConstructors :: Ren QName -> TCM (Ren QName)
     closeConstructors rd = do
-        ds <- nubOn id . catMaybes <$> traverse constructorData (Map.keys rd)
-        cs <- nubOn id . concat    <$> traverse dataConstructors (Map.keys rd)
+        ds <- nubOn snd . catMaybes <$> traverse constructorData (Map.toList rd)
+        cs <- nubOn snd . concat    <$> traverse dataConstructors (Map.toList rd)
         new <- Map.unionsWith (<>) <$> traverse rename (ds ++ cs)
         reportSDoc "tc.mod.apply.complete" 30 $
           "also copying: " <+> pretty new
         return $ Map.unionWith (<>) new rd
       where
-        rename :: QName -> TCM (Ren QName)
-        rename x
+        rename :: (ModuleName, QName) -> TCM (Ren QName)
+        rename (m, x)
           | x `Map.member` rd = pure mempty
           | otherwise =
-              Map.singleton x . pure . qnameFromList . singleton <$> freshName_ (prettyShow $ qnameName x)
+              -- Ulf, 2024-06-24 (#7329):
+              --   Here we used to generate an unqualified name, but this breaks things if the new
+              --   module shows up in a module application later on. This is because we use the
+              --   module name to figure out which arguments from the application are relevant for
+              --   the current symbol (see argsToUse in applySection' below).
+              --
+              --   Instead we use the target module name of the thing that required x to be copied.
+              --   For instance, if we are copying a data type A.B.D to X.Y.Z.D and its constructor
+              --   mkD is not in the renaming, we add `A.B.mkD -> X.Y.Z.mkD` (instead of `A.B.mkD ->
+              --   mkD` which we did before).
+              Map.singleton x . pure . qualify m <$> freshName_ (prettyShow $ qnameName x)
 
-        constructorData :: QName -> TCM (Maybe QName)
-        constructorData x = do
+        constructorData :: (QName, List1.List1 QName) -> TCM (Maybe (ModuleName, QName))
+        constructorData (x, y List1.:| _) = do  -- All new names share the same module, so we can safely grab the first one
           (theDef <$> getConstInfo x) <&> \case
-            Constructor{ conData = d } -> Just d
+            Constructor{ conData = d } -> Just (qnameModule y, d)
             _                          -> Nothing
 
-        dataConstructors :: QName -> TCM [QName]
-        dataConstructors x = do
+        dataConstructors :: (QName, List1.List1 QName) -> TCM [(ModuleName, QName)]
+        dataConstructors (x, y List1.:| _) = do
           (theDef <$> getConstInfo x) <&> \case
-            Datatype{ dataCons = cs } -> cs
-            Record{ recConHead = h }  -> [conName h]
+            Datatype{ dataCons = cs } -> map (qnameModule y,) cs
+            Record{ recConHead = h }  -> [(qnameModule y, conName h)]
             _                         -> []
 
 applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> ScopeCopyInfo -> TCM ()
@@ -502,6 +503,8 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
     -- equivalent valid names and either can be used.
     copyName x = maybe x List1.head (Map.lookup x rd)
 
+    copyConHead c = c { conName = copyName (conName c) }
+
     argsToUse x = do
       let m = commonParentModule old x
       reportSDoc "tc.mod.apply" 80 $ "Common prefix: " <+> pretty m
@@ -526,7 +529,9 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
       -- definition we get will have signature μ \ Δ → B.  This is only valid
       -- for pure modality systems though.
       let ai = defArgInfo def
-          m = unitModality { modCohesion = getCohesion ai }
+          m = unitModality { modCohesion = getCohesion ai
+                           , modPolarity = getModalPolarity ai
+                           }
       localTC (over eContext (map (mapModality (m `inverseComposeModality`)))) $
         copyDef' ts' np def
       where
@@ -545,7 +550,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
           -- Issue5583: Don't skip constructures, because the original constructor doesn't always
           -- work. For instance if it's only available in an anonymous module generated by
           -- `open import M args`.
-          whenJust inst $ \ c -> addNamedInstance y c
+          whenJust inst $ \_ -> addTypedInstance' False False inst y t
           -- Set display form for the old name if it's not a constructor.
 {- BREAKS fail/Issue478
           -- Andreas, 2012-10-20 and if we are not an anonymous module
@@ -562,7 +567,6 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
             t   = defType d `piApply` ts'
             pol = defPolarity d `apply` ts'
             occ = defArgOccurrences d `apply` ts'
-            gen = defArgGeneralizable d `apply` ts'
             inst = defInstance d
             -- the name is set by the addConstant function
             nd :: QName -> TCM Definition
@@ -576,7 +580,6 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                     , defType           = t
                     , defPolarity       = pol
                     , defArgOccurrences = occ
-                    , defArgGeneralizable = gen
                     , defGeneralizedParams = [] -- This is only needed for type checking data/record defs so no need to copy it.
                     , defDisplay        = []
                     , defMutual         = -1   -- TODO: mutual block?
@@ -632,16 +635,19 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                          , dataClause = Just cl
                          , dataCons   = map copyName cs
                          }
-                Record{ recPars = np, recTel = tel } -> return $
+                Record{ recPars = np, recTel = tel, recConHead = c, recFields = fs } -> return $
                   oldDef { recPars    = np - size ts'
                          , recClause  = Just cl
                          , recTel     = apply tel ts'
+                         , recConHead = copyConHead c
+                         , recFields  = (map . fmap) copyName fs
                          }
-                GeneralizableVar -> return GeneralizableVar
+                GeneralizableVar gv -> return $ GeneralizableVar $ gv `apply` ts'
                 _ -> do
                   (mst, _, cc) <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
                   fun          <- emptyFunctionData
                   let newDef =
+                        set funProj   (oldDef ^. funProj) $
                         set funMacro  (oldDef ^. funMacro) $
                         set funStatic (oldDef ^. funStatic) $
                         set funInline True $
@@ -666,8 +672,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                             Function{funProjection = Right p} -> projDropParsApply p ProjSystem rel ts'
                             _ -> Def x $ map Apply ts'
                         , clauseType        = Just $ defaultArg t
-                        , clauseCatchall    = False
-                        , clauseExact       = Just True
+                        , clauseCatchall    = empty
                         , clauseRecursive   = Just False -- definitely not recursive
                         , clauseUnreachable = Just False -- definitely not unreachable
                         , clauseEllipsis    = NoEllipsis
@@ -724,13 +729,14 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
 -- | Add a display form to a definition (could be in this or imported signature).
 addDisplayForm :: QName -> DisplayForm -> TCM ()
 addDisplayForm x df = do
+  -- Check whether display form is recursive and thus illegal.
+  xs <- chaseDisplayForms df Set.empty
+  if x `Set.member` xs then warning $ InvalidDisplayForm x "it is recursive" else do
   d <- makeOpen df
   let add = updateDefinition x $ \ def -> def{ defDisplay = d : defDisplay def }
   ifM (isLocal x)
     {-then-} (modifySignature add)
     {-else-} (stImportsDisplayForms `modifyTCLens` HMap.insertWith (++) x [d])
-  whenM (hasLoopingDisplayForm x) $
-    typeError . GenericDocError =<< do "Cannot add recursive display form for" <+> pretty x
 
 isLocal :: ReadTCState m => QName -> m Bool
 isLocal x = HMap.member x <$> useR (stSignature . sigDefinitions)
@@ -743,24 +749,50 @@ getDisplayForms q = do
   ifM (isLocal q) (return $ ds ++ ds1 ++ ds2)
                   (return $ ds1 ++ ds ++ ds2)
 
+hasDisplayForms :: (HasConstInfo m, ReadTCState m) => QName -> m Bool
+hasDisplayForms = fmap (not . null) . getDisplayForms
+
 -- | Find all names used (recursively) by display forms of a given name.
-chaseDisplayForms :: QName -> TCM (Set QName)
-chaseDisplayForms q = go Set.empty [q]
-  where
-    go :: Set QName        -- Accumulator.
-       -> [QName]          -- Work list.  TODO: make work set to avoid duplicate chasing?
-       -> TCM (Set QName)
-    go used []       = pure used
-    go used (q : qs) = do
-      let rhs (Display _ _ e) = e   -- Only look at names in the right-hand side (#1870)
-      let notYetUsed x = if x `Set.member` used then Set.empty else Set.singleton x
-      ds <- namesIn' notYetUsed . map (rhs . dget)
-            <$> (getDisplayForms q `catchError_` \ _ -> pure [])  -- might be a pattern synonym
-      go (Set.union ds used) (Set.toList ds ++ qs)
+--
+class ChaseDisplayForms a where
+  chaseDisplayForms ::
+       a                 -- ^ Search this recursively for display form names.
+    -> Set QName         -- ^ Already processed names (accumulator).
+    -> TCM (Set QName)   -- ^ Found names (superset of accumulator)
+
+instance ChaseDisplayForms QName where
+  chaseDisplayForms q used
+    | q `Set.member` used = return used
+    | otherwise           = do
+        reportSDoc "tc.display.recursive" 90 $ sep
+          [ "Chasing display form", prettyTCM q, "with accumulator", prettyTCM (Set.toList used) ]
+        xs <- getDisplayForms q `catchError_` const (pure [])  -- might be a pattern synonym
+        chaseDisplayForms xs (Set.insert q used)
+
+instance ChaseDisplayForms DisplayTerm where
+  chaseDisplayForms e used = do
+    let notYetUsed x = if x `Set.member` used then Set.empty else Set.singleton x
+    let ds = namesIn' notYetUsed e
+    chaseDisplayForms ds used
+
+instance ChaseDisplayForms DisplayForm where
+  -- Only look at names in the right-hand side (#1870)
+  chaseDisplayForms = chaseDisplayForms . dfRHS
+
+instance ChaseDisplayForms a => ChaseDisplayForms (Open a) where
+  chaseDisplayForms = chaseDisplayForms . openThing
+
+instance ChaseDisplayForms a => ChaseDisplayForms (Set a) where
+  chaseDisplayForms s = case Set.minView s of
+    Nothing      -> return
+    Just (x, s') -> chaseDisplayForms x >=> chaseDisplayForms s'
+
+instance ChaseDisplayForms a => ChaseDisplayForms [a] where
+  chaseDisplayForms []     = return
+  chaseDisplayForms (x:xs) = chaseDisplayForms x >=> chaseDisplayForms xs
+  -- NB: The following does not work because of lacking instance Ord LocalDisplayForm:
+  -- chaseDisplayForms = chaseDisplayForms . Set.toList
 
--- | Check if a display form is looping.
-hasLoopingDisplayForm :: QName -> TCM Bool
-hasLoopingDisplayForm q = Set.member q <$> chaseDisplayForms q
 
 canonicalName :: HasConstInfo m => QName -> m QName
 canonicalName x = do
@@ -806,7 +838,7 @@ data SigError
     -- Agda, but the current language is Erased Cubical Agda, and
     -- @--erasure@ is not active.
 
--- | Generates an error message corresponding to
+-- | Generates a string error message corresponding to
 -- 'SigCubicalNotErasure' for a given 'QName'.
 notSoPrettySigCubicalNotErasure :: QName -> String
 notSoPrettySigCubicalNotErasure q =
@@ -825,15 +857,15 @@ prettySigCubicalNotErasure q = fsep $
 
 -- | An eliminator for 'SigError'. All constructors except for
 -- 'SigAbstract' are assumed to be impossible.
-sigError :: (HasCallStack, MonadDebug m) => m a -> SigError -> m a
-sigError a = \case
+sigError :: (HasCallStack, MonadDebug m) => QName -> m a -> SigError -> m a
+sigError q a = \case
   SigUnknown s         -> __IMPOSSIBLE_VERBOSE__ s
   SigAbstract          -> a
-  SigCubicalNotErasure -> __IMPOSSIBLE__
+  SigCubicalNotErasure -> __IMPOSSIBLE_VERBOSE__ $
+    notSoPrettySigCubicalNotErasure q
 
 class ( Functor m
       , Applicative m
-      , Fail.MonadFail m
       , HasOptions m
       , MonadDebug m
       , MonadTCEnv m
@@ -841,13 +873,12 @@ class ( Functor m
   -- | Lookup the definition of a name. The result is a closed thing, all free
   --   variables have been abstracted over.
   getConstInfo :: QName -> m Definition
-  getConstInfo q = getConstInfo' q >>= \case
+  getConstInfo q =
+    getConstInfo' q >>= \case
       Right d -> return d
-      Left (SigUnknown err) -> __IMPOSSIBLE_VERBOSE__ err
-      Left SigAbstract      -> __IMPOSSIBLE_VERBOSE__ $
+      Left err -> flip (sigError q) err $
+        __IMPOSSIBLE_VERBOSE__ $
         "Abstract, thus, not in scope: " ++ prettyShow q
-      Left SigCubicalNotErasure -> __IMPOSSIBLE_VERBOSE__ $
-        notSoPrettySigCubicalNotErasure q
 
   -- | Version that reports exceptions:
   getConstInfo' :: QName -> m (Either SigError Definition)
@@ -870,11 +901,11 @@ class ( Functor m
 
 {-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
 
+{-# SPECIALIZE getOriginalConstInfo :: QName -> TCM Definition #-}
 -- | The computation 'getConstInfo' sometimes tweaks the returned
 -- 'Definition', depending on the current 'Language' and the
 -- 'Language' of the 'Definition'. This variant of 'getConstInfo' does
 -- not perform any tweaks.
-
 getOriginalConstInfo ::
   (ReadTCState m, HasConstInfo m) => QName -> m Definition
 getOriginalConstInfo q = do
@@ -891,8 +922,8 @@ getOriginalConstInfo q = do
 defaultGetRewriteRulesFor :: (ReadTCState m, MonadTCEnv m) => QName -> m RewriteRules
 defaultGetRewriteRulesFor q = ifNotM (shouldReduceDef q) (return []) $ do
   st <- getTCState
-  let sig = st^.stSignature
-      imp = st^.stImports
+  let sig = st ^. stSignature
+      imp = st ^. stImports
       look s = HMap.lookup q $ s ^. sigRewriteRules
   return $ mconcat $ catMaybes [look sig, look imp]
 
@@ -911,15 +942,14 @@ instance HasConstInfo (TCMT IO) where
       Right d -> return d
       Left (SigUnknown err)     -> fail err
       Left SigAbstract          -> notInScopeError $ qnameToConcrete q
-      Left SigCubicalNotErasure ->
-        typeError . GenericDocError =<< prettySigCubicalNotErasure q
+      Left SigCubicalNotErasure -> typeError $ CubicalNotErasure q
 
 defaultGetConstInfo
   :: (HasOptions m, MonadDebug m, MonadTCEnv m)
   => TCState -> TCEnv -> QName -> m (Either SigError Definition)
 defaultGetConstInfo st env q = do
-    let defs  = st^.(stSignature . sigDefinitions)
-        idefs = st^.(stImports . sigDefinitions)
+    let defs  = st ^. stSignature . sigDefinitions
+        idefs = st ^. stImports   . sigDefinitions
     case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
         []  -> return $ Left $ SigUnknown $ "Unbound name: " ++ prettyShow q ++ showQNameId q
         [d] -> checkErasureFixQuantity d >>= \case
@@ -1003,13 +1033,34 @@ setPolarity q pol = do
 getForcedArgs :: HasConstInfo m => QName -> m [IsForced]
 getForcedArgs q = defForced <$> getConstInfo q
 
+-- | Returns the occurences given explicitely as polarity annotations in the function type
+getOccurrencesFromType :: Type -> TCM [Occurrence]
+getOccurrencesFromType t = do
+  polarityEnabled <- optPolarity <$> pragmaOptions
+  if polarityEnabled then do
+    telList <- telToList . theTel <$> telView t
+    return $ modalPolarityToOccurrence . modPolarityAnn . getModalPolarity <$> telList
+  else return []
+
 -- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
 getArgOccurrence :: QName -> Nat -> TCM Occurrence
 getArgOccurrence d i = do
   def <- getConstInfo d
-  return $! case theDef def of
-    Constructor{} -> StrictPos
-    _             -> fromMaybe Mixed $ defArgOccurrences def !!! i
+  case theDef def of
+    Constructor{} -> return StrictPos
+
+    -- If the polarity checker has filled in the positivities of
+    -- arguments for this function then we believe it, because it should
+    -- agree with the signature.
+    _ | Just pol <- defArgOccurrences def !!! i ->
+      return pol
+
+    -- Otherwise we use them from the type. Since telView can do
+    -- reduction we'd rather not do this very often.
+    _ -> do
+      occs <- getOccurrencesFromType (defType def)
+
+      pure $! fromMaybe Mixed $ occs !!! i
 
 -- | Sets the 'defArgOccurrences' for the given identifier (which
 -- should already exist in the signature).
@@ -1083,6 +1134,11 @@ getMutual_ = \case
     _ -> Nothing
 
 -- | Set the mutually recursive identifiers.
+--
+--   TODO: This produces data of quadratic size (which has to be processed upon serialization).
+--   Presumably qs is usually short, but in some cases (for instance for generated code) it may be
+--   long. It would be better to assign a unique identifier to each SCC, and store the names
+--   separately.
 setMutual :: QName -> [QName] -> TCM ()
 setMutual d m = modifySignature $ updateDefinition d $ updateTheDef $ \ def ->
   case def of
@@ -1229,11 +1285,11 @@ instantiateDef
 instantiateDef d = do
   vs  <- freeVarsToApply $ defName d
   verboseS "tc.sig.inst" 30 $ do
-    ctx <- getContext
+    ctx <- getContextNames
     m   <- currentModule
     reportSDoc "tc.sig.inst" 30 $
       "instDef in" <+> pretty m <> ":" <+> pretty (defName d) <+>
-      fsep (map pretty $ zipWith (<$) (reverse $ map (fst . unDom) ctx) vs)
+      fsep (map pretty $ zipWith (<$) ctx vs)
   return $ d `apply` vs
 
 instantiateRewriteRule :: (Functor m, HasConstInfo m, HasOptions m,
@@ -1350,7 +1406,7 @@ droppedPars d = case theDef d of
     Constructor{conPars = n} -> n
     Primitive{}              -> 0
     PrimitiveSort{}          -> 0
-    AbstractDefn{}           -> __IMPOSSIBLE__
+    AbstractDefn{}           -> 0 -- not impossible when quoting, PR #7828
 
 -- | Is it the name of a record projection?
 {-# SPECIALIZE isProjection :: QName -> TCM (Maybe Projection) #-}
@@ -1392,7 +1448,7 @@ projectionArgs :: Definition -> Int
 projectionArgs = maybe 0 (max 0 . pred . projIndex) . isRelevantProjection_
 
 -- | Check whether a definition uses copatterns.
-usesCopatterns :: (HasConstInfo m) => QName -> m Bool
+usesCopatterns :: (HasConstInfo m, HasBuiltins m) => QName -> m Bool
 usesCopatterns q = defCopatternLHS <$> getConstInfo q
 
 -- | Apply a function @f@ to its first argument, producing the proper
