@@ -5,7 +5,6 @@ module Agda.TypeChecking.Reduce
  ( Instantiate, instantiate', instantiate, instantiateWhen
  -- Recursive meta instantiation
  , InstantiateFull, instantiateFull', instantiateFull
- , instantiateFullExceptForDefinitions
  -- Check for meta (no reduction)
  , IsMeta, isMeta
  -- Reduction and blocking
@@ -25,8 +24,7 @@ module Agda.TypeChecking.Reduce
  , slowNormaliseArgs
  ) where
 
-import Control.Monad ( (>=>), void )
-import Control.Monad.Except
+import Control.Monad.Except ( MonadError(..) )
 
 import Data.List ( intercalate )
 import Data.Maybe
@@ -41,6 +39,7 @@ import Agda.Interaction.Options
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
+import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.MetaVars
 import Agda.Syntax.Scope.Base (Scope)
@@ -64,9 +63,9 @@ import {-# SOURCE #-} Agda.TypeChecking.Opacity
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.List1 (List1)
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
-import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import qualified Agda.Utils.SmallSet as SmallSet
@@ -91,9 +90,11 @@ instantiateWhen p =
   localR (\env -> env { redPred = Just p }) .
   instantiateFull'
 
+{-# INLINE reduce #-}
 reduce :: (Reduce a, MonadReduce m) => a -> m a
 reduce = liftReduce . reduce'
 
+{-# INLINE reduceB #-}
 reduceB :: (Reduce a, MonadReduce m) => a -> m (Blocked a)
 reduceB = liftReduce . reduceB'
 
@@ -104,6 +105,7 @@ reduceWithBlocker a = ifBlocked a
   (\b a' -> return (b, a'))
   (\_ a' -> return (neverUnblock, a'))
 
+{-# INLINE normalise #-}
 normalise :: (Normalise a, MonadReduce m) => a -> m a
 normalise = liftReduce . normalise'
 
@@ -113,6 +115,7 @@ normalise = liftReduce . normalise'
 -- normaliseB :: (MonadReduce m, Reduce t, Normalise t) => t -> m (Blocked t)
 -- normaliseB = normalise >=> reduceB
 
+{-# INLINE simplify #-}
 simplify :: (Simplify a, MonadReduce m) => a -> m a
 simplify = liftReduce . simplify'
 
@@ -124,6 +127,7 @@ isFullyInstantiatedMeta m = do
     InstV inst -> noMetas <$> instantiateFull (instBody inst)
     _ -> return False
 
+{-# INLINABLE blockAll #-}
 -- | Blocking on all blockers.
 blockAll :: (Functor f, Foldable f) => f (Blocked a) -> Blocked (f a)
 blockAll bs = blockedOn block $ fmap ignoreBlocking bs
@@ -131,6 +135,7 @@ blockAll bs = blockedOn block $ fmap ignoreBlocking bs
         blocker NotBlocked{}  = alwaysUnblock
         blocker (Blocked b _) = b
 
+{-# INLINABLE blockAny #-}
 -- | Blocking on any blockers.
 blockAny :: (Functor f, Foldable f) => f (Blocked a) -> Blocked (f a)
 blockAny bs = blockedOn block $ fmap ignoreBlocking bs
@@ -140,6 +145,7 @@ blockAny bs = blockedOn block $ fmap ignoreBlocking bs
         blocker NotBlocked{}  = []
         blocker (Blocked b _) = [b]
 
+{-# SPECIALIZE blockOnError :: Blocker -> TCM a -> TCM a #-}
 -- | Run the given computation but turn any errors into blocked computations with the given blocker
 blockOnError :: MonadError TCErr m => Blocker -> m a -> m a
 blockOnError blocker f
@@ -147,8 +153,9 @@ blockOnError blocker f
   | otherwise               = f `catchError` \case
     TypeError{}         -> throwError $ PatternErr blocker
     PatternErr blocker' -> throwError $ PatternErr $ unblockOnEither blocker blocker'
-    err@Exception{}     -> throwError err
+    GenericException{}  -> __IMPOSSIBLE__
     err@IOException{}   -> throwError err
+    ParserError{}       -> __IMPOSSIBLE__
 
 -- | Instantiate something.
 --   Results in an open meta variable or a non meta.
@@ -161,6 +168,7 @@ class Instantiate t where
   instantiate' = traverse instantiate'
 
 instance Instantiate t => Instantiate [t]
+instance Instantiate t => Instantiate (List1 t)
 instance Instantiate t => Instantiate (Map k t)
 instance Instantiate t => Instantiate (Maybe t)
 instance Instantiate t => Instantiate (Strict.Maybe t)
@@ -206,9 +214,7 @@ instance Instantiate Term where
          _ | Just m' <- mvTwin mv, blocking ->
            instantiate' (MetaV m' es)
 
-         Open -> return t
-
-         OpenInstance -> return t
+         OpenMeta _ -> return t
 
          BlockedConst u
            | blocking  -> instantiate' . unBrave $
@@ -296,6 +302,7 @@ instance Instantiate Constraint where
   instantiate' (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> instantiate' (a,b)
   instantiate' (UnBlock m)          = return $ UnBlock m
   instantiate' (FindInstance m cs)  = FindInstance m <$> mapM instantiate' cs
+  instantiate' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   instantiate' (IsEmpty r t)        = IsEmpty r <$> instantiate' t
   instantiate' (CheckSizeLtSat t)   = CheckSizeLtSat <$> instantiate' t
   instantiate' c@CheckFunDef{}      = return c
@@ -407,11 +414,14 @@ class Reduce t where
   reduce'  t = ignoreBlocking <$> reduceB' t
   reduceB' t = notBlocked <$> reduce' t
 
+  {-# MINIMAL reduce' | reduceB' #-}
+
 instance Reduce Type where
     reduce'  (El s t) = workOnTypes $ El s <$> reduce' t
     reduceB' (El s t) = workOnTypes $ fmap (El s) <$> reduceB' t
 
 instance Reduce Sort where
+    -- Does not return a 'NotBlocked' 'PiSort', 'FunSort', or 'UnivSort'.
     reduceB' s = do
       s <- instantiate' s
       let done | MetaS x _ <- s = return $ blocked x s
@@ -479,9 +489,9 @@ instance Reduce t => Reduce (Maybe t) where
 
 instance Reduce t => Reduce (Arg t) where
     reduce' a = case getRelevance a of
-      Irrelevant -> return a             -- Don't reduce' irr. args!?
-                                         -- Andreas, 2018-03-03, caused #2989.
-      _          -> traverse reduce' a
+      Irrelevant{} -> return a             -- Don't reduce' irr. args!?
+                                           -- Andreas, 2018-03-03, caused #2989.
+      _ -> traverse reduce' a
 
     reduceB' t = traverse id <$> traverse reduceB' t
 
@@ -807,7 +817,7 @@ reduceHead v = do -- ignoreAbstractMode $ do
       abstractMode <- envAbstractMode <$> askTC
       isAbstract <- not <$> hasAccessibleDef f
       traceSLn "tc.inj.reduce" 50 (
-        "reduceHead: we are in " ++ show abstractMode++ "; " ++ prettyShow f ++
+        "reduceHead: we are in " ++ show abstractMode ++ "; " ++ prettyShow f ++
         " is treated " ++ if isAbstract then "abstractly" else "concretely"
         ) $ do
       let v0  = Def f []
@@ -827,7 +837,8 @@ reduceHead v = do -- ignoreAbstractMode $ do
         _                               -> return $ notBlocked v
     _ -> return $ notBlocked v
 
--- | Unfold a single inlined function.
+-- | Unfold as many copies as possible, and then potentially a single
+-- inline function.
 unfoldInlined :: PureTCM m => Term -> m Term
 unfoldInlined v = do
   inTypes <- viewTC eWorkingOnTypes
@@ -835,8 +846,14 @@ unfoldInlined v = do
     _ | inTypes -> return v -- Don't inline in types (to avoid unfolding of goals)
     Def f es -> do
       info <- getConstInfo f
-      let def = theDef info
-          irr = isIrrelevant $ defArgInfo info
+
+      let
+        def = theDef info
+        irr = isIrrelevant $ defArgInfo info
+        continue
+          | defCopy info = fmap notBlocked . unfoldInlined
+          | otherwise    = return . notBlocked
+
       case def of
         Function{} ->
           reportSLn "tc.inline" 90 $
@@ -847,12 +864,13 @@ unfoldInlined v = do
             , "funCompiled = " ++ prettyShow (funCompiled def)
             ]
         _ -> pure ()
-      case def of   -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
+
+      case def of -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
         Function{ funCompiled = Just Done{} }
-          | def ^. funInline , not irr -> do
-              reportSLn "tc.inline" 70 $ "asking to inline " ++ prettyShow f
-              liftReduce $
-                ignoreBlocking <$> unfoldDefinitionE (return . notBlocked) (Def f []) f es
+          | (defCopy info || def ^. funInline), not irr -> do
+            reportSLn "tc.inline" 70 $ "asking to inline " ++ prettyShow f
+            liftReduce $
+              ignoreBlocking <$> unfoldDefinitionE continue (Def f []) f es
         _ -> return v
     _ -> return v
 
@@ -914,13 +932,23 @@ appDefE'' v cls rewr es = traceSDoc "tc.reduce" 90 ("appDefE' v = " <+> pretty v
               nvars = size $ clauseTel cl
           -- if clause is underapplied, skip to next clause
           if length es < npats then goCls cls es else do
+            termCheck <- asksTC envTermCheckReducing
             let (es0, es1) = splitAt npats es
             (m, es0) <- matchCopatterns pats es0
             let es = es0 ++ es1
             case m of
-              No         -> goCls cls es
-              DontKnow b -> rewrite b (applyE v) rewr es
+              No               -> goCls cls es
+              -- Szumi, 2024-03-29, issue #7181:
+              -- If a lazy match is stuck and all non-lazy matches are conclusive,
+              -- then reduction should not be stuck on the current clause and it
+              -- should be fine to continue matching on the next clause.
+              -- This assumes it's impossible for a lazy match to be stuck if
+              -- all non-lazy matches succeed.
+              DontKnow OnlyLazy _ -> goCls cls es
+              DontKnow NonLazy  b -> rewrite b (applyE v) rewr es
               Yes simpl vs -- vs is the subst. for the variables bound in body
+                | termCheck && fromMaybe True (clauseRecursive cl) ->
+                    return $ NoReduction __IMPOSSIBLE__
                 | Just w <- body -> do -- clause has body?
                     -- TODO: let matchPatterns also return the reduced forms
                     -- of the original arguments!
@@ -933,6 +961,7 @@ instance Reduce a => Reduce (Closure a) where
     reduce' cl = do
         x <- enterClosure cl reduce'
         return $ cl { clValue = x }
+{-# SPECIALIZE reduce' :: Closure Constraint -> ReduceM (Closure Constraint) #-}
 
 instance Reduce Telescope where
   reduce' EmptyTel          = return EmptyTel
@@ -951,6 +980,7 @@ instance Reduce Constraint where
   reduce' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> reduce' (a,b)
   reduce' (UnBlock m)           = return $ UnBlock m
   reduce' (FindInstance m cs)   = FindInstance m <$> mapM reduce' cs
+  reduce' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   reduce' (IsEmpty r t)         = IsEmpty r <$> reduce' t
   reduce' (CheckSizeLtSat t)    = CheckSizeLtSat <$> reduce' t
   reduce' c@CheckFunDef{}       = return c
@@ -1117,6 +1147,7 @@ instance Simplify Constraint where
   simplify' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> simplify' (a,b)
   simplify' (UnBlock m)           = return $ UnBlock m
   simplify' (FindInstance m cs)   = FindInstance m <$> mapM simplify' cs
+  simplify' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   simplify' (IsEmpty r t)         = IsEmpty r <$> simplify' t
   simplify' (CheckSizeLtSat t)    = CheckSizeLtSat <$> simplify' t
   simplify' c@CheckFunDef{}       = return c
@@ -1177,9 +1208,10 @@ class Normalise t where
   default normalise' :: (t ~ f a, Traversable f, Normalise a) => t -> ReduceM t
   normalise' = traverse normalise'
 
--- boring instances:
+-- Functor instances:
 
 instance Normalise t => Normalise [t]
+instance Normalise t => Normalise (List1 t)
 instance Normalise t => Normalise (Map k t)
 instance Normalise t => Normalise (Maybe t)
 instance Normalise t => Normalise (Strict.Maybe t)
@@ -1190,6 +1222,8 @@ instance Normalise t => Normalise (Named name t)
 instance Normalise t => Normalise (IPBoundary' t)
 instance Normalise t => Normalise (WithHiding t)
 
+-- more boring instances:
+
 instance (Normalise a, Normalise b) => Normalise (a,b) where
     normalise' (x,y) = (,) <$> normalise' x <*> normalise' y
 
@@ -1298,6 +1332,7 @@ instance Normalise Constraint where
   normalise' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> normalise' (a,b)
   normalise' (UnBlock m)           = return $ UnBlock m
   normalise' (FindInstance m cs)   = FindInstance m <$> mapM normalise' cs
+  normalise' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   normalise' (IsEmpty r t)         = IsEmpty r <$> normalise' t
   normalise' (CheckSizeLtSat t)    = CheckSizeLtSat <$> normalise' t
   normalise' c@CheckFunDef{}       = return c
@@ -1363,6 +1398,7 @@ class InstantiateFull t where
 -- Traversables (doesn't include binders like Abs, Tele):
 
 instance InstantiateFull t => InstantiateFull [t]
+instance InstantiateFull t => InstantiateFull (List1 t)
 instance InstantiateFull t => InstantiateFull (HashMap k t)
 instance InstantiateFull t => InstantiateFull (Map k t)
 instance InstantiateFull t => InstantiateFull (Maybe t)
@@ -1498,6 +1534,9 @@ instance (Subst a, InstantiateFull a) => InstantiateFull (Abs a) where
 instance (InstantiateFull t, InstantiateFull e) => InstantiateFull (Dom' t e) where
     instantiateFull' (Dom i n b tac x) = Dom i n b <$> instantiateFull' tac <*> instantiateFull' x
 
+instance InstantiateFull ContextEntry where
+  instantiateFull' (CtxVar x a) = CtxVar x <$> instantiateFull' a
+
 instance InstantiateFull LetBinding where
   instantiateFull' (LetBinding o v t) = LetBinding o <$> instantiateFull' v <*> instantiateFull' t
 
@@ -1538,6 +1577,7 @@ instance InstantiateFull Constraint where
     SortCmp cmp a b     -> uncurry (SortCmp cmp) <$> instantiateFull' (a,b)
     UnBlock m           -> return $ UnBlock m
     FindInstance m cs   -> FindInstance m <$> mapM instantiateFull' cs
+    ResolveInstanceHead q -> return $ ResolveInstanceHead q
     IsEmpty r t         -> IsEmpty r <$> instantiateFull' t
     CheckSizeLtSat t    -> CheckSizeLtSat <$> instantiateFull' t
     c@CheckFunDef{}     -> return c
@@ -1557,7 +1597,11 @@ instance InstantiateFull CompareAs where
   instantiateFull' AsTypes       = return AsTypes
 
 instance InstantiateFull Signature where
-  instantiateFull' (Sig a b c) = uncurry3 Sig <$> instantiateFull' (a, b, c)
+  instantiateFull' (Sig a b c d) = Sig
+    <$> instantiateFull' a
+    <*> instantiateFull' b
+    <*> instantiateFull' c
+    <*> pure d             -- The instance table only stores names
 
 instance InstantiateFull Section where
   instantiateFull' (Section tel) = Section <$> instantiateFull' tel
@@ -1665,13 +1709,12 @@ instance InstantiateFull CompiledClauses where
   instantiateFull' (Case n bs) = Case n <$> instantiateFull' bs
 
 instance InstantiateFull Clause where
-    instantiateFull' (Clause rl rf tel ps b t catchall exact recursive unreachable ell wm) =
+    instantiateFull' (Clause rl rf tel ps b t catchall recursive unreachable ell wm) =
        Clause rl rf <$> instantiateFull' tel
        <*> instantiateFull' ps
        <*> instantiateFull' b
        <*> instantiateFull' t
        <*> return catchall
-       <*> return exact
        <*> return recursive
        <*> return unreachable
        <*> return ell
@@ -1694,47 +1737,27 @@ instance InstantiateFull RemoteMetaVariable where
     <*> instantiateFull' c
 
 instance InstantiateFull Interface where
-  instantiateFull' i = do
-    defs <- instantiateFull' (i ^. intSignature . sigDefinitions)
-    instantiateFullExceptForDefinitions'
-      (set (intSignature . sigDefinitions) defs i)
-
--- | Instantiates everything except for definitions in the signature.
-
-instantiateFullExceptForDefinitions' :: Interface -> ReduceM Interface
-instantiateFullExceptForDefinitions'
-  (Interface h s ft ms mod tlmod scope inside sig metas display userwarn
-     importwarn b foreignCode highlighting libPragmas filePragmas
-     usedOpts patsyns warnings partialdefs oblocks onames) =
-  Interface h s ft ms mod tlmod scope inside
-    <$> ((\s r -> Sig { _sigSections     = s
-                      , _sigDefinitions  = sig ^. sigDefinitions
-                      , _sigRewriteRules = r
-                      })
-         <$> instantiateFull' (sig ^. sigSections)
-         <*> instantiateFull' (sig ^. sigRewriteRules))
-    <*> instantiateFull' metas
-    <*> instantiateFull' display
-    <*> return userwarn
-    <*> return importwarn
-    <*> instantiateFull' b
-    <*> return foreignCode
-    <*> return highlighting
-    <*> return libPragmas
-    <*> return filePragmas
-    <*> return usedOpts
-    <*> return patsyns
-    <*> return warnings
-    <*> return partialdefs
-    <*> return oblocks
-    <*> return onames
-
--- | Instantiates everything except for definitions in the signature.
-
-instantiateFullExceptForDefinitions ::
-  MonadReduce m => Interface -> m Interface
-instantiateFullExceptForDefinitions =
-  liftReduce . instantiateFullExceptForDefinitions'
+  instantiateFull'
+    (Interface h s ft ms mod tlmod scope inside sig _ display userwarn
+         importwarn b foreignCode highlighting libPragmas filePragmas
+         usedOpts patsyns warnings partialdefs oblocks onames) = do
+    Interface h s ft ms mod tlmod scope inside
+      <$!> instantiateFull' sig
+      <*!> pure mempty               -- remote metas are dropped
+      <*!> instantiateFull' display
+      <*!> return userwarn
+      <*!> return importwarn
+      <*!> instantiateFull' b
+      <*!> return foreignCode
+      <*!> return highlighting
+      <*!> return libPragmas
+      <*!> return filePragmas
+      <*!> return usedOpts
+      <*!> return patsyns
+      <*!> return warnings
+      <*!> return partialdefs
+      <*!> return oblocks
+      <*!> return onames
 
 instance InstantiateFull a => InstantiateFull (Builtin a) where
     instantiateFull' (Builtin t) = Builtin <$> instantiateFull' t
