@@ -126,10 +126,9 @@ module Agda.TypeChecking.Rules.LHS.Unify
 
 import Prelude hiding (null)
 
-import Control.Monad
-import Control.Monad.State
-import Control.Monad.Writer (WriterT(..), MonadWriter(..))
-import Control.Monad.Except
+import Control.Monad.State  ( gets, modify, evalStateT )
+import Control.Monad.Writer ( WriterT(..), MonadWriter(..) )
+import Control.Monad.Except ( runExceptT )
 
 import Data.Semigroup hiding (Arg)
 import qualified Data.List as List
@@ -146,7 +145,7 @@ import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
-import Agda.TypeChecking.Conversion.Pure
+import Agda.TypeChecking.Conversion.Pure (pureEqualTermB, pureEqualTypeB)
 import Agda.TypeChecking.Constraints ()
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Irrelevance
@@ -164,7 +163,6 @@ import Agda.TypeChecking.Rules.LHS.Problem
 import Agda.TypeChecking.Rules.LHS.Unify.Types
 import Agda.TypeChecking.Rules.LHS.Unify.LeftInverse
 
-import Agda.Utils.Benchmark
 import Agda.Utils.Either
 import Agda.Utils.Function
 import Agda.Utils.Functor
@@ -213,27 +211,25 @@ data UnificationResult' a
 --
 --   The result is the most general unifier of @us@ and @vs@.
 unifyIndices
-  :: (PureTCM m, MonadBench m, BenchPhase m ~ Bench.Phase, MonadError TCErr m)
-  => Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
+  :: Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
   -> Telescope       -- ^ @gamma@
   -> FlexibleVars    -- ^ @flex@
   -> Type            -- ^ @a@
   -> Args            -- ^ @us@
   -> Args            -- ^ @vs@
-  -> m UnificationResult
+  -> TCM UnificationResult
 unifyIndices linv tel flex a us vs =
   Bench.billTo [Bench.Typing, Bench.CheckLHS, Bench.UnifyIndices] $
     fmap (\(a,b,c,_) -> (a,b,c)) <$> unifyIndices' linv tel flex a us vs
 
 unifyIndices'
-  :: (PureTCM m, MonadError TCErr m)
-  => Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
+  :: Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
   -> Telescope     -- ^ @gamma@
   -> FlexibleVars  -- ^ @flex@
   -> Type          -- ^ @a@
   -> Args          -- ^ @us@
   -> Args          -- ^ @vs@
-  -> m FullUnificationResult
+  -> TCM FullUnificationResult
 unifyIndices' linv tel flex a [] [] = return $ Unifies (tel, idS, [], Right (idS, raiseS 1))
 unifyIndices' linv tel flex a us vs = do
     reportSDoc "tc.lhs.unify" 10 $
@@ -266,7 +262,7 @@ unifyIndices' linv tel flex a us vs = do
 
 
 
-type UnifyStrategy = forall m. (PureTCM m, MonadPlus m) => UnifyState -> m UnifyStep
+type UnifyStrategy = UnifyState -> ListT TCM UnifyStep
 
 
 --UNUSED Liang-Ting Chen 2019-07-16
@@ -525,11 +521,10 @@ skipIrrelevantStrategy k s = do
 ----------------------------------------------------
 
 unifyStep
-  :: (PureTCM m, MonadWriter UnifyOutput m, MonadError TCErr m)
-  => UnifyState -> UnifyStep -> m (UnificationResult' UnifyState)
+  :: UnifyState -> UnifyStep -> UnifyStepT TCM (UnificationResult' UnifyState)
 unifyStep s Deletion{ deleteAt = k , deleteType = a , deleteLeft = u , deleteRight = v } = do
     -- Check definitional equality of u and v
-    isReflexive <- addContext (varTel s) $ runBlocked $ pureEqualTerm a u v
+    isReflexive <- addContext (varTel s) $ pureEqualTermB a u v
     withoutK <- withoutKOption
     splitOnStrict <- asksTC envSplitOnStrict
     case isReflexive of
@@ -582,7 +577,7 @@ unifyStep s (Injectivity k a d pars ixs c) = do
   -- a left inverse for the overall match, so as a slight optimisation
   -- we just don't bother computing it. __IMPOSSIBLE__ because that
   -- field in the result is never evaluated.
-  res <- addContext (varTel s) $ unifyIndices' (Just __IMPOSSIBLE__)
+  res <- lift $ addContext (varTel s) $ unifyIndices' (Just __IMPOSSIBLE__)
            hduTel
            (allFlexVars notforced hduTel)
            (raise (size ctel) dtype)
@@ -683,11 +678,11 @@ unifyStep s Cycle
 
 unifyStep s EtaExpandVar{ expandVar = fi, expandVarRecordType = d , expandVarParameters = pars } = do
   recd <- fromMaybe __IMPOSSIBLE__ <$> isRecord d
-  let delta = recTel recd `apply` pars
-      c     = recConHead recd
+  let delta = _recTel recd `apply` pars
+      c     = _recConHead recd
   let nfields         = size delta
       (varTel', rho)  = expandTelescopeVar (varTel s) (m-1-i) delta c
-      projectFlexible = [ FlexibleVar (getArgInfo fi) (flexForced fi) (projFlexKind j) (flexPos fi) (i+j) | j <- [0..nfields-1] ]
+      projectFlexible = [ FlexibleVar (getArgInfo fi) (flexForced fi) (projFlexKind j) (flexPos fi) (i + j) | j <- [0 .. nfields - 1] ]
   tellUnifySubst $ rho
   return $ Unifies $ UState
     { varTel   = varTel'
@@ -715,8 +710,8 @@ unifyStep s EtaExpandVar{ expandVar = fi, expandVarRecordType = d , expandVarPar
 
 unifyStep s EtaExpandEquation{ expandAt = k, expandRecordType = d, expandParameters = pars } = do
   recd  <- fromMaybe __IMPOSSIBLE__ <$> isRecord d
-  let delta = recTel recd `apply` pars
-      c     = recConHead recd
+  let delta = _recTel recd `apply` pars
+      c     = _recConHead recd
   lhs   <- expandKth $ eqLHS s
   rhs   <- expandKth $ eqRHS s
   let (tel, sigma) = expandTelescopeVar (eqTel s) k delta c
@@ -830,10 +825,10 @@ solutionStep retry s
   -- Check that the type of the variable is equal to the type of the equation
   -- (not just a subtype), otherwise we cannot instantiate (see Issue 2407).
   let dom'@Dom{ unDom = a' } = getVarType (m-1-i) s
-  equalTypes <- addContext (varTel s) $ runBlocked $ do
+  equalTypes <- addContext (varTel s) $ do
     reportSDoc "tc.lhs.unify" 45 $ "Equation type: " <+> prettyTCM a
     reportSDoc "tc.lhs.unify" 45 $ "Variable type: " <+> prettyTCM a'
-    pureEqualType a a'
+    pureEqualTypeB a a'
 
   -- The conditions on the relevances are as follows (see #2640):
   -- - If the type of the equation is relevant, then the solution must be
@@ -849,7 +844,7 @@ solutionStep retry s
   let eqrel  = getRelevance dom
       eqmod  = getModality dom
       varmod = getModality dom'
-      mod    = applyUnless (NonStrict `moreRelevant` eqrel) (setRelevance eqrel)
+      mod    = applyUnless (shapeIrrelevant `moreRelevant` eqrel) (setRelevance eqrel)
              $ applyUnless (usableQuantity envmod) (setQuantity zeroQuantity)
              $ varmod
   reportSDoc "tc.lhs.unify" 65 $ text $ "Equation modality: " ++ show (getModality dom)
@@ -893,32 +888,28 @@ solutionStep retry s
     Right True -> return $ UnifyStuck [UnifyUnusableModality (varTel s) a i u mod]
 solutionStep _ _ _ = __IMPOSSIBLE__
 
-unify
-  :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
-  => UnifyState -> UnifyStrategy -> m (UnificationResult' UnifyState)
+unify :: UnifyState -> UnifyStrategy -> UnifyLogT TCM (UnificationResult' UnifyState)
 unify s strategy = if isUnifyStateSolved s
                    then return $ Unifies s
                    else tryUnifyStepsAndContinue (strategy s)
   where
     tryUnifyStepsAndContinue
-      :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
-      => ListT m UnifyStep -> m (UnificationResult' UnifyState)
+      :: ListT TCM UnifyStep -> UnifyLogT TCM (UnificationResult' UnifyState)
     tryUnifyStepsAndContinue steps = do
-      x <- foldListT tryUnifyStep failure steps
+      x <- foldListT tryUnifyStep failure (liftListT lift steps)
       case x of
         Unifies s'     -> unify s' strategy
         NoUnify err    -> return $ NoUnify err
         UnifyBlocked b -> return $ UnifyBlocked b
         UnifyStuck err -> return $ UnifyStuck err
 
-    tryUnifyStep :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
-                 => UnifyStep
-                 -> m (UnificationResult' UnifyState)
-                 -> m (UnificationResult' UnifyState)
+    tryUnifyStep :: UnifyStep
+                 -> UnifyLogT TCM (UnificationResult' UnifyState)
+                 -> UnifyLogT TCM (UnificationResult' UnifyState)
     tryUnifyStep step fallback = do
       addContext (varTel s) $
         reportSDoc "tc.lhs.unify" 20 $ "trying unifyStep" <+> prettyTCM step
-      (x, output) <- runWriterT $ unifyStep s step
+      (x, output) <- lift $ runWriterT $ unifyStep s step
       case x of
         Unifies s'   -> do
           reportSDoc "tc.lhs.unify" 20 $ "unifyStep successful."
