@@ -4,6 +4,7 @@
 module Agda.TypeChecking.Serialise.Instances.Internal where
 
 import qualified Data.HashSet as HashSet
+import Control.Monad
 import Control.Monad.IO.Class
 
 import Agda.Syntax.Internal as I
@@ -16,7 +17,9 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Coverage.SplitTree
+import Agda.TypeChecking.DiscrimTree.Types
 
+import Agda.Utils.Functor
 import Agda.Utils.Permutation
 
 import Agda.Utils.Impossible
@@ -27,10 +30,15 @@ instance EmbPrj a => EmbPrj (Dom a) where
   value = valueN Dom
 
 instance EmbPrj Signature where
-  icod_ (Sig a b c) = icodeN' Sig a b c
+  icod_ (Sig a b c d) = icodeN' Sig a b c d
 
   value = valueN Sig
 
+instance EmbPrj InstanceTable where
+  icod_ (InstanceTable a b) = icodeN' InstanceTable a b
+
+  value = valueN InstanceTable
+
 instance EmbPrj Section where
   icod_ (Section a) = icodeN' Section a
 
@@ -186,7 +194,7 @@ instance EmbPrj a => EmbPrj (Open a) where
 
 instance EmbPrj CheckpointId where
   icod_ (CheckpointId a) = icode a
-  value n                = CheckpointId `fmap` value n
+  value n                = CheckpointId <$!> value n
 
 instance EmbPrj DisplayTerm where
   icod_ (DTerm'   a b)   = icodeN' DTerm' a b
@@ -204,16 +212,29 @@ instance EmbPrj DisplayTerm where
     valu _            = malformed
 
 instance EmbPrj MutualId where
-  icod_ (MutId a) = icode a
-  value n         = MutId `fmap` value n
+  icod_ (MutualId a) = icode a
+  value n            = MutualId <$!> value n
 
 instance EmbPrj CompKit where
   icod_ (CompKit a b) = icodeN' CompKit a b
   value = valueN CompKit
 
-instance EmbPrj Definition where
-  icod_ (Defn a b c d e f g h i j k l m n o p q r s) = icodeN' Defn a b (P.killRange c) d e f g h i j k l m n o p q r s
+instance EmbPrj InstanceInfo where
+  icod_ (InstanceInfo a b) = icodeN' InstanceInfo a b
+  value = valueN InstanceInfo
 
+instance EmbPrj Definition where
+  icod_ (Defn a b c d e f g h i j k l m n o blocked r s) =
+    icodeN' Defn a b (P.killRange c) d e f g h i j k l m n o (ossify blocked) r s
+    where
+      -- Andreas, 2024-01-02, issue #7044:
+      -- After serialization, a definition can never be unblocked,
+      -- since all metas are ossified.
+      -- Thus, we turn any blocker into 'neverUnblock'.
+      ossify :: Blocked_ -> Blocked_
+      ossify = \case
+        b@NotBlocked{} -> b
+        Blocked b () -> Blocked neverUnblock ()
   value = valueN Defn
 
 instance EmbPrj NotBlocked where
@@ -231,11 +252,22 @@ instance EmbPrj NotBlocked where
     valu [3, a] = valuN MissingClauses a
     valu _      = malformed
 
-instance EmbPrj Blocked_ where
-  icod_ (NotBlocked a b) = icodeN' NotBlocked a b
-  icod_ Blocked{} = __IMPOSSIBLE__
+-- Andreas, 2024-01-02, issue #7044.
+-- We only serialize 'neverUnblock';
+-- other than that, there should not be any blockers left at serialization time.
+blockedToMaybe :: Blocked_ -> Maybe NotBlocked
+blockedToMaybe = \case
+  NotBlocked a ()       -> Just a
+  Blocked a ()
+    | a == neverUnblock -> Nothing
+    | otherwise         -> __IMPOSSIBLE__
 
-  value = valueN NotBlocked
+blockedFromMaybe :: Maybe NotBlocked -> Blocked_
+blockedFromMaybe = maybe (Blocked neverUnblock ()) (`NotBlocked` ())
+
+instance EmbPrj Blocked_ where
+  icod_ = icod_ . blockedToMaybe
+  value = blockedFromMaybe <.> value
 
 instance EmbPrj NLPat where
   icod_ (PVar a b)      = icodeN 0 PVar a b
@@ -389,26 +421,29 @@ instance EmbPrj BuiltinSort where
 
 instance EmbPrj Defn where
   icod_ (Axiom       a)                                 = icodeN 0 Axiom a
-  icod_ (Function    a b s t u c d e f g h i j k l m)   = icodeN 1 (\ a b s -> Function a b s t) a b s u c d e f g h i j k l m
+  icod_ (Function    a b s t u c d e f g h i j k)       = icodeN 1 (\ a b s -> Function a b s t) a b s u c d e f g h i j k
   icod_ (Datatype    a b c d e f g h i j)               = icodeN 2 Datatype a b c d e f g h i j
   icod_ (Record      a b c d e f g h i j k l m)         = icodeN 3 Record a b c d e f g h i j k l m
   icod_ (Constructor a b c d e f g h i j k)             = icodeN 4 Constructor a b c d e f g h i j k
   icod_ (Primitive   a b c d e f)                       = icodeN 5 Primitive a b c d e f
   icod_ (PrimitiveSort a b)                             = icodeN 6 PrimitiveSort a b
   icod_ AbstractDefn{}                                  = __IMPOSSIBLE__
-  icod_ GeneralizableVar                                = icodeN 7 GeneralizableVar
-  icod_ DataOrRecSig{}                                  = __IMPOSSIBLE__
+  icod_ (GeneralizableVar a)                            = icodeN 7 GeneralizableVar a
+  icod_ (DataOrRecSig a)                                = icodeN 8 DataOrRecSig a
+    -- Andreas, 2024-10-27
+    -- DataOrRecSig is possible via unquoteDecl in meta-programming, see #7576
 
   value = vcase valu where
     valu [0, a]                                        = valuN Axiom a
-    valu [1, a, b, s, u, c, d, e, f, g, h, i, j, k, l, m]
-                                                       = valuN (\ a b s -> Function a b s Nothing) a b s u c d e f g h i j k l m
+    valu [1, a, b, s, u, c, d, e, f, g, h, i, j, k]
+                                                       = valuN (\ a b s -> Function a b s Nothing) a b s u c d e f g h i j k
     valu [2, a, b, c, d, e, f, g, h, i, j]             = valuN Datatype a b c d e f g h i j
     valu [3, a, b, c, d, e, f, g, h, i, j, k, l, m]    = valuN Record   a b c d e f g h i j k l m
     valu [4, a, b, c, d, e, f, g, h, i, j, k]          = valuN Constructor a b c d e f g h i j k
     valu [5, a, b, c, d, e, f]                         = valuN Primitive   a b c d e f
     valu [6, a, b]                                     = valuN PrimitiveSort a b
-    valu [7]                                           = valuN GeneralizableVar
+    valu [7, a]                                        = valuN GeneralizableVar a
+    valu [8, a]                                        = valuN DataOrRecSig a
     valu _                                             = malformed
 
 instance EmbPrj LazySplit where
@@ -440,16 +475,7 @@ instance EmbPrj a => EmbPrj (SplitTree' a) where
     valu [0, a, b, c] = valuN SplitAt a b c
     valu _            = malformed
 
-instance EmbPrj FunctionFlag where
-  icod_ FunStatic       = pure 0
-  icod_ FunInline       = pure 1
-  icod_ FunMacro        = pure 2
-
-  value = \case
-    0 -> pure FunStatic
-    1 -> pure FunInline
-    2 -> pure FunMacro
-    _ -> malformed
+instance EmbPrj FunctionFlag
 
 instance EmbPrj a => EmbPrj (WithArity a) where
   icod_ (WithArity a b) = icodeN' WithArity a b
@@ -468,10 +494,12 @@ instance EmbPrj a => EmbPrj (Case a) where
 instance EmbPrj OpaqueBlock where
   icod_ (OpaqueBlock id uf _ _ r) =
     icodeN' (\id uf ->
-      OpaqueBlock id (HashSet.fromList uf) mempty Nothing)
+      let !unfolding = HashSet.fromList uf
+      in OpaqueBlock id unfolding mempty Nothing)
     id (HashSet.toList uf) r
 
-  value = valueN (\id uf -> OpaqueBlock id (HashSet.fromList uf) mempty Nothing)
+  value = valueN (\id uf -> let !unfolding = HashSet.fromList uf
+                            in OpaqueBlock id unfolding mempty Nothing)
 
 instance EmbPrj CompiledClauses where
   icod_ (Fail a)   = icodeN' Fail a
@@ -509,7 +537,7 @@ instance EmbPrj TermHead where
     valu _      = malformed
 
 instance EmbPrj I.Clause where
-  icod_ (Clause a b c d e f g h i j k l) = icodeN' Clause a b c d e f g h i j k l
+  icod_ (Clause a b c d e f g h i j k) = icodeN' Clause a b c d e f g h i j k
 
   value = valueN Clause
 
@@ -538,6 +566,7 @@ instance EmbPrj I.PatOrigin where
   icod_ PatORec     = icodeN 6 PatORec
   icod_ PatOLit     = icodeN 7 PatOLit
   icod_ PatOAbsurd  = icodeN 8 PatOAbsurd
+  icod_ (PatOSplitArg a) = icodeN 9 PatOSplitArg a
 
   value = vcase valu where
     valu []     = valuN PatOSystem
@@ -549,6 +578,7 @@ instance EmbPrj I.PatOrigin where
     valu [6]    = valuN PatORec
     valu [7]    = valuN PatOLit
     valu [8]    = valuN PatOAbsurd
+    valu [9, a] = valuN PatOSplitArg a
     valu _      = malformed
 
 instance EmbPrj a => EmbPrj (I.Pattern' a) where
@@ -625,3 +655,31 @@ instance EmbPrj a => EmbPrj (Judgement a) where
 instance EmbPrj RemoteMetaVariable where
   icod_ (RemoteMetaVariable a b c) = icodeN' RemoteMetaVariable a b c
   value = valueN RemoteMetaVariable
+
+instance EmbPrj Key where
+  icod_ (RigidK x y) = icodeN 0 RigidK x y
+  icod_ (LocalK x y) = icodeN 1 LocalK x y
+  icod_ (PiK h)      = icodeN 2 PiK h
+  icod_ FlexK        = icodeN 3 FlexK
+  icod_ ConstK       = icodeN 4 ConstK
+  icod_ SortK        = icodeN 5 SortK
+
+  value = vcase valu where
+    valu [0, x, y] = valuN RigidK x y
+    valu [1, x, y] = valuN LocalK x y
+    valu [2, h]    = valuN PiK h
+    valu [3]       = valuN FlexK
+    valu [4]       = valuN ConstK
+    valu [5]       = valuN SortK
+    valu _         = malformed
+
+instance (EmbPrj a, Ord a) => EmbPrj (DiscrimTree a) where
+  icod_ EmptyDT        = icodeN' EmptyDT
+  icod_ (DoneDT s)     = icodeN' DoneDT s
+  icod_ (CaseDT i k s) = icodeN' CaseDT i k s
+
+  value = vcase valu where
+    valu []        = valuN EmptyDT
+    valu [a]       = valuN DoneDT a
+    valu [i, k, s] = valuN CaseDT i k s
+    valu _         = malformed
